"""
XIA Slit from EPICS synApps optics: xia_slit.db
+++++++++++++++++++++++++++++++++++++++++++++++

Coordinates (viewing from detector towards source)::

        top
    inb     out
        bot

Each blade [#]_ (in the XIA slit controller) travels in a _cylindrical_
coordinate system.  Positive motion moves a blade **outwards** from the center
with a backlash correction. No backlash correction is applied for negative
motion (as the blades close).  Size and center are computed
by the underlying EPICS support.

    hsize = inb + out
    vsize = top + bot

..  [#] Note that the blade names here are different than the EPICS support.
    The difference is to make the names of the blades consistent with other
    slits with the Bluesky framework.

USAGE:

    slit = XiaSlitController("IOC:hsc1:", name="slit")
    print(slit.geometry)

.. autosummary::

   ~XiaSlit2D
"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ..devices import PVPositionerSoftDone
from ophyd import Signal

from ..utils import SlitGeometry


class XiaSlit2D(Device):
    """
    EPICS synApps optics xia_slit.db 2D support: inb out bot top ...
    """

    inb = Component(
        PVPositionerSoftDone, "", setpoint_pv="l", readback_pv="lRB"
    )
    out = Component(
        PVPositionerSoftDone, "", setpoint_pv="r", readback_pv="rRB"
    )
    bot = Component(
        PVPositionerSoftDone, "", setpoint_pv="b", readback_pv="bRB"
    )
    top = Component(
        PVPositionerSoftDone, "", setpoint_pv="t", readback_pv="tRB"
    )

    hsize = Component(
        PVPositionerSoftDone, "", setpoint_pv="width", readback_pv="widthRB"
    )
    vsize = Component(
        PVPositionerSoftDone, "", setpoint_pv="height", readback_pv="heightRB"
    )
    hcenter = Component(
        PVPositionerSoftDone, "", setpoint_pv="h0", readback_pv="h0RB"
    )
    vcenter = Component(
        PVPositionerSoftDone, "", setpoint_pv="v0", readback_pv="v0RB"
    )

    hID = Component(EpicsSignal, "hID", kind="config", string=True)
    horientation = Component(EpicsSignal, "hOrient", kind="config", string=True)
    hbusy = Component(EpicsSignalRO, "hBusy", kind="omitted", string=True)

    vID = Component(EpicsSignal, "vID", kind="config", string=True)
    vorientation = Component(EpicsSignal, "vOrient", kind="config", string=True)
    vbusy = Component(EpicsSignalRO, "vBusy", kind="omitted", string=True)

    enable = Component(EpicsSignal, "enable", kind="config", string=True)

    # diagnostics
    error_code = Component(EpicsSignalRO, "err", kind="config")
    error_message = Component(EpicsSignalRO, "errMsg", kind="config", string=True)
    message1 = Component(EpicsSignalRO, "msg1", kind="config", string=True)
    message2 = Component(EpicsSignalRO, "msg2", kind="config", string=True)
    message3 = Component(EpicsSignalRO, "msg3", kind="config", string=True)

    # configuration
    calibrate = Component(EpicsSignal, "calib", kind="omitted", put_complete=True)
    initialize = Component(EpicsSignal, "init", kind="omitted", put_complete=True)
    locate = Component(EpicsSignal, "locate", kind="omitted", put_complete=True)
    stop_button = Component(EpicsSignal, "stop", kind="omitted", put_complete=True)

    precision = Component(Signal, value=3, kind="config")

    @property
    def geometry(self):
        """Return the slit 2D size and center as a namedtuple."""
        pppp = [
            round(obj.position, self.precision.get())
            for obj in (
                self.hsize,
                self.vsize,
                self.hcenter,
                self.vcenter
            )
        ]

        return SlitGeometry(*pppp)

    @geometry.setter
    def geometry(self, value):
        # first, test the input by assigning it to local vars
        width, height, x, y = value

        self.hsize.move(width)
        self.vsize.move(height)
        self.hcenter.move(x)
        self.vcenter.move(y)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Ophyd support for the EPICS synApps swait record

EXAMPLES::

    import apstools.synApps
    calcs = apstools.synApps.UserCalcsDevice("xxx:", name="calcs")

    calc1 = calcs.calc1
    apstools.synApps.setup_random_number_swait(calc1)

    apstools.synApps.setup_incrementer_swait(calcs.calc2)

    calc1.reset()


.. autosummary::

    ~UserCalcN
    ~UserCalcsDevice
    ~SwaitRecord
    ~SwaitRecordChannel
    ~setup_random_number_swait
    ~setup_gaussian_swait
    ~setup_lorentzian_swait
    ~setup_incrementer_swait

:see: https://htmlpreview.github.io/?https://raw.githubusercontent.com/epics-modules/calc/R3-6-1/documentation/swaitRecord.html
"""


from collections import OrderedDict
from ophyd import Device
from ophyd import Component as Cpt
from ophyd import DynamicDeviceComponent as DDC
from ophyd import FormattedComponent as FC
from ophyd import EpicsSignal
from ophyd.signal import EpicsSignalBase

from ._common import EpicsRecordDeviceCommonAll
from ._common import EpicsSynAppsRecordEnableMixin
from .. import utils as APS_utils


CHANNEL_LETTERS_LIST = "A B C D E F G H I J K L".split()


class SwaitRecordChannel(Device):
    """
    EPICS synApps synApps swait record: single channel [A-L]

    .. index:: Ophyd Device; synApps SwaitRecordChannel
    """

    input_value = FC(EpicsSignal, "{self.prefix}.{self._ch_letter}", kind="config")
    input_pv = FC(EpicsSignal, "{self.prefix}.IN{self._ch_letter}N", kind="config")
    input_trigger = FC(EpicsSignal, "{self.prefix}.IN{self._ch_letter}P", kind="config")
    read_attrs = [
        "input_value",
    ]
    hints = {"fields": read_attrs}

    def __init__(self, prefix, letter, **kwargs):
        self._ch_letter = letter
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.input_pv.put("")
        self.input_value.put(0)
        self.input_trigger.put("Yes")


def _swait_channels(channel_list):
    defn = OrderedDict()
    for chan in channel_list:
        defn[chan] = (SwaitRecordChannel, "", {"letter": chan})
    return defn


class SwaitRecord(EpicsRecordDeviceCommonAll):
    """
    EPICS synApps swait record: used as ``$(P):userCalc$(N)``

    .. index:: Ophyd Device; synApps SwaitRecord

    .. autosummary::

        ~reset

    """
    precision = Cpt(EpicsSignal, ".PREC", kind="config")
    high_operating_range = Cpt(EpicsSignal, ".HOPR", kind="config")
    low_operating_range = Cpt(EpicsSignal, ".LOPR", kind="config")

    calculated_value = Cpt(EpicsSignal, ".VAL", kind="hinted")
    calculation = Cpt(EpicsSignal, ".CALC", kind="config")

    output_link_pv = Cpt(EpicsSignal, ".OUTN", kind="config")
    output_location_name = Cpt(EpicsSignal, ".DOLN", kind="config")
    output_location_data = Cpt(EpicsSignal, ".DOLD", kind="config")
    output_data_option = Cpt(EpicsSignal, ".DOPT", kind="config")

    output_execute_option = Cpt(EpicsSignal, ".OOPT", kind="config")
    output_execution_delay = Cpt(EpicsSignal, ".ODLY", kind="config")
    event_to_issue = Cpt(EpicsSignal, ".OEVT", kind="config")

    read_attrs = APS_utils.itemizer("channels.%s", CHANNEL_LETTERS_LIST)
    hints = {"fields": read_attrs}

    channels = DDC(_swait_channels(CHANNEL_LETTERS_LIST))

    @property
    def value(self):
        return self.calculated_value.get()

    def reset(self):
        """set all fields to default values"""
        pvname = self.description.pvname.split(".")[0]
        self.description.put(pvname)
        self.scanning_rate.put("Passive")
        self.calculation.put("0")
        self.precision.put("5")
        self.output_location_data.put(0)
        self.output_location_name.put("")
        self.output_data_option.put("Use VAL")
        self.forward_link.put("0")
        self.output_execution_delay.put(0)
        self.output_execute_option.put("Every Time")
        self.output_link_pv.put("")
        for letter in self.channels.read_attrs:
            channel = getattr(self.channels, letter)
            if isinstance(channel, SwaitRecordChannel):
                channel.reset()

        self.read_attrs = ["channels.%s" % c for c in CHANNEL_LETTERS_LIST]
        self.hints = {"fields": self.read_attrs}

        self.read_attrs.append("calculated_value")


class UserCalcN(EpicsSynAppsRecordEnableMixin, SwaitRecord):
    """Single instance of the userCalcN database."""


class UserCalcsDevice(Device):
    """
    EPICS synApps XXX IOC setup of userCalcs: ``$(P):userCalc$(N)``

    .. index:: Ophyd Device; synApps UserCalcsDevice

    .. autosummary::

        ~reset

    """

    enable = Cpt(EpicsSignal, "userCalcEnable", kind="config")
    calc1 = Cpt(UserCalcN, "userCalc1")
    calc2 = Cpt(UserCalcN, "userCalc2")
    calc3 = Cpt(UserCalcN, "userCalc3")
    calc4 = Cpt(UserCalcN, "userCalc4")
    calc5 = Cpt(UserCalcN, "userCalc5")
    calc6 = Cpt(UserCalcN, "userCalc6")
    calc7 = Cpt(UserCalcN, "userCalc7")
    calc8 = Cpt(UserCalcN, "userCalc8")
    calc9 = Cpt(UserCalcN, "userCalc9")
    calc10 = Cpt(UserCalcN, "userCalc10")

    def reset(self):  # lgtm [py/similar-function]
        """set all fields to default values"""
        self.calc1.reset()
        self.calc2.reset()
        self.calc3.reset()
        self.calc4.reset()
        self.calc5.reset()
        self.calc6.reset()
        self.calc7.reset()
        self.calc8.reset()
        self.calc9.reset()
        self.calc10.reset()
        self.read_attrs = ["calc%d" % (c + 1) for c in range(10)]


def setup_random_number_swait(swait, **kw):
    """setup swait record to generate random numbers"""
    swait.reset()
    swait.scanning_rate.put("Passive")
    swait.description.put("uniform random numbers")
    swait.calculation.put("RNDM")
    swait.scanning_rate.put(".1 second")

    swait.read_attrs = [
        "calculated_value",
    ]
    swait.hints = {"fields": swait.read_attrs}


def _setup_peak_swait_(
    calc, desc, swait, ref_signal, center=0, width=1, scale=1, noise=0.05
):
    """
    internal: setup that is common to both Gaussian and Lorentzian swaits

    PARAMETERS

    swait
        *object* :
        instance of :class:`SwaitRecord`

    ref_signal
        *object* :
        instance of :class:`EpicsSignal` used as $A$

    center
        *float* :
        EPICS record field $B$,
        default = 0

    width
        *float* :
        EPICS record field $C$,
        default = 1

    scale
        *float* :
        EPICS record field $D$,
        default = 1

    noise
        *float* :
        EPICS record field $E$,
        default = 0.05
    """

    # consider a noisy background, as well (needs a couple calcs)
    if not isinstance(swait, SwaitRecord):
        raise TypeError("expected SwaitRecord instance," f" received {type(swait)}")
    if not isinstance(ref_signal, EpicsSignalBase):
        raise TypeError(
            "expected EpicsSignalBase instance," f" received {type(ref_signal)}"
        )
    if width <= 0:
        raise ValueError(f"width must be positive, received {width}")
    if not (0.0 <= noise <= 1.0):
        raise ValueError(f"noise must be between 0 and 1, received {noise}")

    swait.reset()
    swait.scanning_rate.put("Passive")
    swait.description.put(desc)
    swait.channels.A.input_pv.put(ref_signal.pvname)
    swait.channels.B.input_value.put(center)
    swait.channels.C.input_value.put(width)
    swait.channels.D.input_value.put(scale)
    swait.channels.E.input_value.put(noise)
    swait.calculation.put(calc)
    swait.scanning_rate.put("I/O Intr")

    swait.read_attrs = [
        "calculated_value",
    ]
    swait.hints = {"fields": swait.read_attrs}


def setup_gaussian_swait(swait, ref_signal, center=0, width=1, scale=1, noise=0.05):
    """
    setup swait for noisy Gaussian

    calculation: $D*(0.95+E*RNDM)/exp(((A-B)/C)^2)$

    PARAMETERS

    swait
        *object* :
        instance of :class:`SwaitRecord`

    ref_signal
        *object* :
        instance of :class:`EpicsSignal` used as $A$

    center
        *float* :
        instance of :class:`EpicsSignal` used as $B$,
        default = 0

    width
        *float* :
        instance of :class:`EpicsSignal` used as $C$,
        default = 1

    scale
        *float* :
        instance of :class:`EpicsSignal` used as $D$,
        default = 1

    noise
        *float* :
        instance of :class:`EpicsSignal` used as $E$,
        default = 0.05
    """
    _setup_peak_swait_(
        "D*(0.95+E*RNDM)/exp(((A-b)/c)^2)",
        "noisy Gaussian curve",
        swait,
        ref_signal,
        center=center,
        width=width,
        scale=scale,
        noise=noise,
    )


def setup_lorentzian_swait(swait, ref_signal, center=0, width=1, scale=1, noise=0.05):
    """
    setup swait record for noisy Lorentzian

    calculation: $D*(0.95+E*RNDM)/(1+((A-B)/C)^2)$

    PARAMETERS

    swait
        *object* :
        instance of :class:`SwaitRecord`

    ref_signal
        *object* :
        instance of :class:`EpicsSignal` used as $A$

    center
        *float* :
        instance of :class:`EpicsSignal` used as $B$,
        default = 0

    width
        *float* :
        instance of :class:`EpicsSignal` used as $C$,
        default = 1

    scale
        *float* :
        instance of :class:`EpicsSignal` used as $D$,
        default = 1

    noise
        *float* :
        instance of :class:`EpicsSignal` used as $E$,
        default = 0.05
    """
    _setup_peak_swait_(
        "D*(0.95+E*RNDM)/(1+((A-b)/c)^2)",
        "noisy Lorentzian curve",
        swait,
        ref_signal,
        center=center,
        width=width,
        scale=scale,
        noise=noise,
    )


def setup_incrementer_swait(swait, scan=None, limit=100000):
    """
    setup swait record as an incrementer

    PARAMETERS

    swait
        *object* :
        instance of :class:`SwaitRecord`

    scan
        *text* or *int* or ``None``
        any of the EPICS record ``.SCAN`` values,
        or the index number of the value,
        set to default if ``None``,
        default: ``.1 second``

    limit
        *int* or ``None``
        set the incrementer back to zero
        when this number is reached (or passed),
        default: 100000

    """
    # consider a noisy background, as well (needs a couple calcs)
    scan = scan or ".1 second"
    swait.reset()
    swait.description.put("incrementer")
    swait.scanning_rate.put("Passive")
    pvname = swait.calculated_value.pvname
    swait.channels.A.input_pv.put(pvname)
    swait.channels.B.input_value.put(limit)
    swait.calculation.put("(A+1) % B")
    swait.precision.put(0)
    swait.scanning_rate.put(scan)

    swait.read_attrs = [
        "calculated_value",
    ]
    swait.hints = {"fields": swait.read_attrs}

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
test the APS cycle computation code
"""

import datetime
import pytest
import socket

from .. import ApsCycleDM
from .. import aps_cycle


def using_APS_workstation():
    hostname = socket.gethostname()
    return hostname.lower().endswith(".aps.anl.gov")


def test_ApsCycleDM():
    signal = ApsCycleDM(name="signal")
    assert signal.connected

    cycle = signal.get()  # expect 2021-3 or such
    assert isinstance(cycle, str)
    assert cycle != ""
    assert len(cycle) == 6
    assert cycle.startswith("20")
    assert cycle.find("-") >= 0


def test_ApsCycleDB():
    assert aps_cycle.YAML_CYCLE_FILE.exists()

    cycle = aps_cycle.cycle_db.get_cycle_name()
    assert cycle in aps_cycle.cycle_db.db
    assert isinstance(cycle, str)
    assert cycle != ""
    assert len(cycle) == 6
    assert cycle.startswith("20")
    assert cycle.find("-") >= 0
    assert int(cycle.split("-")[0]) > 2020


@pytest.mark.parametrize(
    "iso8601, cycle_name",
    [
        ["2014-06-28 01:23:45", "2014-2"],
        ["2021-09-28 12:34:56", "2021-2"],
        ["2021-09-30 23:59:59", "2021-2"],
        ["2021-10-02 00:00:00", "2021-3"],
        ["1999-06-01", None],
    ],
)
def test_cycles(iso8601, cycle_name):
    assert isinstance(iso8601, str)

    dt = datetime.datetime.fromisoformat(iso8601)
    ts = datetime.datetime.timestamp(dt)
    assert aps_cycle.cycle_db.get_cycle_name(ts) == cycle_name



import pytest
from ...migration import spec2ophyd
import pathlib
from collections import OrderedDict


path = pathlib.Path(__file__).parent.parent


def test_Version():
    from ... import __version__
    version = spec2ophyd.get_apstools_version()
    assert isinstance(version, str)
    assert len(version) > len("#.#.#")
    assert version == __version__


def test_SpecConfig():
    sc = spec2ophyd.SpecConfig(path / "config_fourc")
    assert sc is not None
    assert isinstance(sc.devices, OrderedDict)
    assert isinstance(sc.scalers, list)
    assert isinstance(sc.collection, list)
    assert isinstance(sc.unhandled, list)
    assert sc.devices == {}
    assert sc.scalers == []
    assert sc.collection == []
    assert sc.unhandled == []

    # no exceptions yet
    sc = spec2ophyd.SpecConfig("cannot find this file")
    assert sc is not None
    assert isinstance(sc.devices, OrderedDict)


def test_SpecConfig_read_config():
    sc = spec2ophyd.SpecConfig("cannot find this file")
    # now the exception
    with pytest.raises(FileNotFoundError) as exinfo:
        sc.read_config()
    assert "No such file or directory:" in str(exinfo.value)

    sc = spec2ophyd.SpecConfig(path / "config")  # USAXS
    sc.read_config()
    assert len(sc.devices) == 3
    assert len(sc.scalers) == 1
    assert len(sc.collection) == 68
    assert len(sc.unhandled) == 2

    sc = spec2ophyd.SpecConfig(path / "config_fourc")
    sc.read_config()
    assert len(sc.devices) == 0
    assert len(sc.scalers) == 0
    assert len(sc.collection) == 7
    assert isinstance(sc.collection[0], spec2ophyd.SpecMotor)
    assert isinstance(sc.collection[1], spec2ophyd.SpecMotor)
    assert isinstance(sc.collection[2], spec2ophyd.SpecMotor)
    assert isinstance(sc.collection[3], spec2ophyd.SpecMotor)
    assert isinstance(sc.collection[4], spec2ophyd.SpecCounter)
    assert isinstance(sc.collection[5], spec2ophyd.SpecCounter)
    assert isinstance(sc.collection[6], spec2ophyd.SpecCounter)
    assert len(sc.unhandled) == 2
    assert sc.unhandled == ["SW_SFTWARE\t = 1 INTR", ""]

    sc = spec2ophyd.SpecConfig(path / "config-CNTPAR")
    sc.read_config()
    assert len(sc.devices) == 4
    assert len(sc.scalers) == 1
    assert len(sc.collection) == 78
    assert len(sc.unhandled) == 7

    sc = spec2ophyd.SpecConfig(path / "config-MOTPAR")
    sc.read_config()
    assert len(sc.devices) == 4
    assert len(sc.scalers) == 1
    assert len(sc.collection) == 124
    assert len(sc.unhandled) == 1

    sc = spec2ophyd.SpecConfig(path / "config_spec")
    sc.read_config()
    assert len(sc.devices) == 3
    assert len(sc.scalers) == 1
    assert len(sc.collection) == 121
    assert len(sc.unhandled) == 12


def test_SpecConfig_find_pv_in_collection():
    sc = spec2ophyd.SpecConfig(path / "config")  # USAXS
    sc.read_config()
    assert isinstance(
        sc.find_pv_in_collection("9idcLAX:aero:c1:m1"),
        spec2ophyd.SpecMotor
    )
    assert isinstance(
        sc.find_pv_in_collection("9idcLAX:vsc:c0"),
        spec2ophyd.SpecSignal
    )
    # FIXME:
    # assert isinstance(
    #     sc.find_pv_in_collection("9idcLAX:vsc:c0.S1"),
    #     spec2ophyd.SpecCounter
    # )


def test_create_ophyd_setup(capsys):
    sc = spec2ophyd.SpecConfig(path / "config_fourc")
    sc.read_config()
    spec2ophyd.create_ophyd_setup(sc)
    out, err = capsys.readouterr()
    lines = out.splitlines()
    assert len(lines) == 22
    assert len(err.splitlines()) == 0
    # TODO:


def test_issue499(capsys):
    """
    failed when parsing this SPEC config line

        MOT022 = MAC_MOT:1/0/0   2000  1  2000  200   50  125    0 0x003   suvgap  SlitUpVGap

    * Only expecting 2 arguments after `MAC_MOT:`  chan/ignore.
    * This seems to be: chan/axis/num
    * only chan is used
    * assign the others into SpecMotor.ctrl_parms as [int]
    """
    sc = spec2ophyd.SpecConfig(path / "config_spec")
    sc.read_config()
    spec2ophyd.create_ophyd_setup(sc)
    out, err = capsys.readouterr()
    lines = out.splitlines()
    assert len(lines) == 137  # this may change if other items are processed
    assert len(err.splitlines()) == 0

    # assert out == "DEBUG"  # enable to show output from create_ophyd_setup(sc)

    # verify the correct configuration line is selected
    m = sc.collection[22]
    assert isinstance(m, spec2ophyd.SpecMotor)
    assert m.mne == "suvgap"
    assert m.name == "SlitUpVGap"
    assert m.motpar == dict(misc_par_1="19", misc_par_2="18")
    assert m.mac_parms == [0, 0]
    assert m.device.prefix == "slit"
    assert m.device.num_channels == 8
    assert m.device.config_line == 15

    # test for the parameters that triggered the issue
    assert sc.collection[22].mac_parms == [0, 0]
    assert sc.collection[23].mac_parms == [0, 1]
    assert sc.collection[24].mac_parms == [1, 0]
    assert sc.collection[25].mac_parms == [1, 1]



import pytest

from ..xia_slit import XiaSlit2D
from ...utils import SlitGeometry
from ...tests import IOC

PV_PREFIX = f"{IOC}phony_hsc1:"

# Full testing requires two XIA Slit controllers (a 2D H & V set)
# We don't have that for unit testing.  Proceed with best efforts.


def test_XiaSlit_not_connected():
    slit1 = XiaSlit2D(PV_PREFIX, name="slit1")
    assert slit1 is not None

    # slit1.wait_for_connection()
    assert not slit1.connected


def test_XiaSlit_geometry(capsys):
    slit1 = XiaSlit2D(PV_PREFIX, name="slit1")

    g = None
    with pytest.raises(TypeError):
        g = slit1.geometry
    assert not isinstance(g, SlitGeometry)

    captured = capsys.readouterr()
    assert captured.out.split("\n") == [""]
    assert captured.err.split("\n") == [""]


def test_XiaSlit_components():
    slit1 = XiaSlit2D(PV_PREFIX, name="slit1")
    cns = """
        inb out bot top
        hsize vsize hcenter vcenter
        hID horientation hbusy
        vID vorientation vbusy
        enable
        error_code error_message message1 message2 message3
        calibrate initialize locate stop_button
        precision
    """.split()
    assert sorted(slit1.component_names) == sorted(cns)



"""
Ophyd support for the EPICS synApps saveData support

see:  https://epics.anl.gov/bcda/synApps/sscan/sscanRecord.html

EXAMPLE::

    from apstools.synApps import SaveData
    save_data = SaveData("xxx:saveData_", name="save_data")


Public Structures

.. autosummary::

    ~SaveData

"""


from ophyd import Device
from ophyd import Component
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO


class SaveData(Device):
    """
    EPICS synApps saveData support.

    .. index:: Ophyd Device; synApps SaveData

    usage::

        from apstools.synApps import SaveData
        save_data = SaveData("xxx:saveData_", name="save_data")

    .. autosummary::

        ~reset

    """

    file_system = Component(EpicsSignal, "fileSystem", kind="config")
    subdirectory = Component(EpicsSignal, "subDir", kind="config")
    base_name = Component(EpicsSignal, "baseName", kind="config")
    next_scan_number = Component(EpicsSignal, "scanNumber", kind="config")
    comment1 = Component(EpicsSignal, "comment1", string=True, kind="config")
    comment2 = Component(EpicsSignal, "comment2", string=True, kind="config")
    write_1D_each_point = Component(
        EpicsSignal, "realTime1D", string=True, kind="config"
    )
    max_retries = Component(EpicsSignal, "maxAllowedRetries", kind="config")
    retry_wait_s = Component(EpicsSignal, "retryWaitinSecs", kind="config")

    full_path_name = Component(
        EpicsSignalRO, "fullPathName", string=True, kind="config"
    )
    full_name = Component(EpicsSignalRO, "fileName", string=True, kind="normal")
    message = Component(EpicsSignalRO, "message", string=True, kind="config")
    status = Component(EpicsSignalRO, "status", string=True, kind="config")
    write_count_current = Component(EpicsSignalRO, "currRetries", kind="config")
    write_count_total = Component(EpicsSignalRO, "totalRetries", kind="config")
    write_count_abandoned = Component(EpicsSignalRO, "abandonedWrites", kind="config")

    def reset(self):
        self.file_system.put("")
        self.subdirectory.put("")
        self.base_name.put("")
        self.next_scan_number.put(1)
        self.comment1.put("")
        self.comment2.put("")
        self.write_1D_each_point.put("No")
        self.max_retries.put(10)
        self.retry_wait_s.put(15)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
file: /tmp/kohzu.py
"""

import time

from .. import KohzuSeqCtl_Monochromator
from ...tests import MASTER_TIMEOUT
from bluesky import plan_stubs as bps
from bluesky import RunEngine
from ophyd import Component
from ophyd import EpicsMotor
from ophyd.signal import EpicsSignalBase


# set default timeout for all EpicsSignal connections & communications
try:
    EpicsSignalBase.set_defaults(
        auto_monitor=True,
        timeout=MASTER_TIMEOUT,
        write_timeout=MASTER_TIMEOUT,
        connection_timeout=MASTER_TIMEOUT,
    )
except RuntimeError:
    pass  # ignore if some EPICS object already created


class MyKohzu(KohzuSeqCtl_Monochromator):
    """
    Kohzu DCM with controls to enable wavelength controls.

    Wavelength controls are disabled when the physical motors are out of range
    for the choice of wavelength (such as mathematical singularity for energy
    when ``m_theta == 0` on start of the IOC.)
    """

    m_theta = Component(EpicsMotor, "m45")
    m_y = Component(EpicsMotor, "m46")
    m_z = Component(EpicsMotor, "m47")

    def into_control_range(self, p_theta=11, p_y=-18, p_z=90):
        """
        Move the Kohzu motors into range so the wavelength controls will work.

        Written as a bluesky plan so that all motors can be moved
        simultaneously. Return early if the motors are already in range.
        """
        args = []
        if self.m_theta.position < p_theta:
            args += [self.m_theta, p_theta]
            yield from bps.mv(self.m_theta.velocity, 5)
        if self.m_y.position > p_y:
            args += [self.m_y, p_y]
            yield from bps.mv(self.m_y.velocity, 5)
        if self.m_z.position < p_z:
            args += [self.m_z, p_z]
            yield from bps.mv(self.m_z.velocity, 5)
        if (len(args) == 0):
            # all motors in range, no work to do, MUST yield something
            yield from bps.null()
            return
        yield from bps.mv(*args)
        yield from bps.sleep(1)  # allow IOC to react
        yield from bps.mv(self.operator_acknowledge, 1)

    def stop(self):
        self.m_theta.stop()
        self.m_y.stop()
        self.m_z.stop()


def test_dcm():
    dcm = MyKohzu("gp:", name="dcm")
    assert dcm is not None
    assert dcm.energy.name == "dcm_energy"
    dcm.wait_for_connection()

    RE = RunEngine({})
    RE(dcm.into_control_range())
    assert dcm.energy.position > 0
    assert round(dcm.m_theta.position, 2) >= 2
    assert round(dcm.m_y.position, 2) <= -15
    assert round(dcm.m_z.position, 2) >= 90

    delay = 0.1
    time.sleep(delay)

    dcm.mode.put("Auto")
    time.sleep(delay)
    assert dcm.mode.get() == "Auto"

    dcm.energy.move(10.2)
    assert round(dcm.energy.setpoint.get(), 7) == 10.2
    assert round(dcm.energy.position, 1) == 10.2

    dcm.energy.move(10.2)  # send it to the same place
    assert round(dcm.energy.position, 1) == 10.2



from .. import make_dict_device
from ...tests import IOC
import ophyd
import pytest
import time


m1 = ophyd.EpicsMotor(f"{IOC}m1", name="m1")
m1.wait_for_connection()


@pytest.mark.parametrize(
    "obj",
    [
        m1.read(),
        dict(a=1, b="two"),
        dict(a="a", b=dict(value="b", timestamp=time.time())),
        dict(a="a", b=dict(values="b", timestamp=time.time())),
        dict(a="a", b=dict(values="b")),
        dict(a="a", b=dict()),
        dict(a="a", b=dict(b=dict(b=dict(b=dict(b=dict()))))),
        dict(a="a", b=dict(value="b", timestamp=time.time(), temperature=25.2, units="any")),
    ],
)
def test_dict_device(obj):
    assert m1.connected

    m1_read = m1.read()
    ddev = make_dict_device(m1_read)
    for k in ddev.component_names:
        assert k in m1_read, k
    for k, v in ddev.read().items():
        k2 = k.lstrip(f"{ddev.name}_")
        assert k2 in m1_read
        assert v == m1_read[k2]



"""
Struck 3820
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~Struck3820
"""

from ophyd import Component
from ophyd import Device
from ophyd.mca import EpicsMCARecord
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO


class Struck3820(Device):
    """
    Struck/SIS 3820 Multi-Channel Scaler (as used by USAXS)

    .. index:: Ophyd Device; Struck3820
    """

    start_all = Component(EpicsSignal, "StartAll")
    stop_all = Component(EpicsSignal, "StopAll")
    erase_start = Component(EpicsSignal, "EraseStart")
    erase_all = Component(EpicsSignal, "EraseAll")
    mca1 = Component(EpicsMCARecord, "mca1")
    mca2 = Component(EpicsMCARecord, "mca2")
    mca3 = Component(EpicsMCARecord, "mca3")
    mca4 = Component(EpicsMCARecord, "mca4")
    clock_frequency = Component(EpicsSignalRO, "clock_frequency")
    current_channel = Component(EpicsSignalRO, "CurrentChannel")
    channel_max = Component(EpicsSignalRO, "MaxChannels")
    channels_used = Component(EpicsSignal, "NuseAll")
    elapsed_real_time = Component(EpicsSignalRO, "ElapsedReal")
    preset_real_time = Component(EpicsSignal, "PresetReal")
    dwell_time = Component(EpicsSignal, "Dwell")
    prescale = Component(EpicsSignal, "Prescale")
    acquiring = Component(EpicsSignalRO, "Acquiring", string=True)
    acquire_mode = Component(EpicsSignalRO, "AcquireMode", string=True)
    model = Component(EpicsSignalRO, "Model", string=True)
    firmware = Component(EpicsSignalRO, "Firmware")
    channel_advance = Component(EpicsSignal, "ChannelAdvance")
    count_on_start = Component(EpicsSignal, "CountOnStart")
    software_channel_advance = Component(EpicsSignal, "SoftwareChannelAdvance")
    channel1_source = Component(EpicsSignal, "Channel1Source")
    user_led = Component(EpicsSignal, "UserLED")
    mux_output = Component(EpicsSignal, "MUXOutput")
    input_mode = Component(EpicsSignal, "InputMode")
    output_mode = Component(EpicsSignal, "OutputMode")
    output_polarity = Component(EpicsSignal, "OutputPolarity")
    read_rate = Component(EpicsSignal, "ReadAll.SCAN")
    do_readl_all = Component(EpicsSignal, "DoReadAll")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



import pytest

from ..luascript import LuascriptRecord
from ..luascript import UserScriptsDevice
from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing


PV_PREFIX = f"{IOC}set1:"
EMPIRICAL_DELAY = 0.5

# TODO: test a luascript file, where to put the file in the IOC?
# LUA_SCRIPT_PATH
# https://epics-lua.readthedocs.io/en/latest/luascriptRecord.html?highlight=file#examples
# https://epics-lua.readthedocs.io/en/latest/using-lua-shell.html?highlight=file#calling-the-lua-shell-from-inside-the-ioc-shell


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [LuascriptRecord, f"{PV_PREFIX}userScript9", False, "read_attrs", 10],
        [LuascriptRecord, f"{PV_PREFIX}userScript9", False, "configuration_attrs", 98],
        [LuascriptRecord, f"{PV_PREFIX}userScript9", True, "read()", 2],
        [LuascriptRecord, f"{PV_PREFIX}userScript9", True, "summary()", 207],

        [UserScriptsDevice, PV_PREFIX, False, "read_attrs", 110],
        [UserScriptsDevice, PV_PREFIX, False, "configuration_attrs", 991],
        [UserScriptsDevice, PV_PREFIX, True, "read()", 20],
        [UserScriptsDevice, PV_PREFIX, True, "summary()", 1906],
    ]
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


def test_luascript_reset():
    lua_all = UserScriptsDevice(PV_PREFIX, name="user")
    lua_all.wait_for_connection()
    lua_all.enable.put("Enable")

    lua = lua_all.script9
    assert isinstance(lua, LuascriptRecord)

    lua_all.reset()
    lua.disable_value.put(2)  # ensure record is always enabled
    short_delay_for_EPICS_IOC_database_processing(EMPIRICAL_DELAY)  # a short-ish wait (discovered empirically)
    assert (
        lua.scan_disable_input_link_value.get() != lua.disable_value.get()
    )

    assert lua.precision.get() == 5
    assert lua.code.get() == ""
    assert lua.number_value.get() == 0
    assert lua.string_value.get() == ""
    assert lua.inputs.A.input_value.get() == 0
    assert lua.inputs.AA.pv_link.get() == ""
    assert lua.inputs.BB.input_value.get() == ""

    lua.precision.put(3)
    lua.number_value.put(-3.7)
    lua.string_value.put("string_value")
    lua.inputs.A.input_value.put(-2)
    lua.inputs.AA.pv_link.put("string_pv_link")
    lua.inputs.BB.input_value.put("BB.input_value")
    # order is important, set this LAST
    lua.code.put("code")
    short_delay_for_EPICS_IOC_database_processing(EMPIRICAL_DELAY)  # a short-ish wait (discovered empirically)

    assert lua.precision.get() != 5
    assert lua.code.get() != ""
    assert lua.number_value.get() != 0
    assert lua.string_value.get() != ""
    assert lua.inputs.A.input_value.get() != 0
    assert lua.inputs.AA.pv_link.get() != ""
    assert lua.inputs.BB.input_value.get() != ""

    lua_all.reset()
    short_delay_for_EPICS_IOC_database_processing()

    assert lua.precision.get() == 5
    assert lua.code.get() == ""
    assert lua.number_value.get() == 0
    assert lua.string_value.get() == ""
    assert lua.inputs.A.input_value.get() == 0
    assert lua.inputs.AA.pv_link.get() == ""
    assert lua.inputs.BB.input_value.get() == ""


@pytest.mark.parametrize(
    "code, a, b, nval, aa, bb, sval",
    [
        ['return "A"', 0, 0, 0, "", "", "A"],
        ["return A+B", 5, -3, 2, "", "", ""],
        ["return A+B  --this is a comment", 5, -3, 2, "", "", ""],
        ['return string.rep(".", 5)', 0, 0, 0, "", "", "....."],
        ['return string.len("parametrize")', 0, 0, 11, "", "", ""],
        ['return "AA" .. "BB"', 0, 0, 0, "", "", "AABB"],

        # TODO: string variables are not recognized in luascript 3.0.1 in the test VM.
        # Wait for the test VM to be upgraded to at least 3.0.2 before enabling these tests.
        # FIXME: ["return string.upper(AA)", 0, 0, 0, "Abc", "", "ABC"],
        # FIXME: ["return string.lower(AA)", 0, 0, 0, "Abc", "", "abc"],
        # FIXME: ["return AA .. BB", 0, 0, 0, "Aaa", "Bbb", "AaaBbb"],
    ]
)
def test_compute(code, a, b, nval, aa, bb, sval):
    lua_all = UserScriptsDevice(PV_PREFIX, name="user")
    lua_all.wait_for_connection()
    lua_all.enable.put("Enable")
    short_delay_for_EPICS_IOC_database_processing()
    assert lua_all.enable.get(as_string=True) == "Enable"

    lua_all.reset()
    short_delay_for_EPICS_IOC_database_processing()

    lua = lua_all.script9
    lua.disable_value.put(2)  # ensure record is always enabled
    assert (
        lua.scan_disable_input_link_value.get() != lua.disable_value.get()
    )

    # set the inputs
    lua.inputs.A.input_value.put(a)
    lua.inputs.B.input_value.put(b)
    lua.inputs.AA.input_value.put(aa)
    lua.inputs.BB.input_value.put(bb)

    # set the lua code
    lua.code.put(code)

    # not computed yet
    assert round(lua.number_value.get(), 5) == 0
    assert lua.string_value.get() == ""

    # do the computation
    lua.process_record.put(1)
    short_delay_for_EPICS_IOC_database_processing(EMPIRICAL_DELAY)  # a short-ish wait (discovered empirically)

    assert round(lua.number_value.get(), 5) == nval
    assert lua.string_value.get() == sval

    lua_all.reset()
    short_delay_for_EPICS_IOC_database_processing()



import pytest

from ..scalcout import ScalcoutRecord
from ..scalcout import UserScalcoutDevice
from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [ScalcoutRecord, f"{IOC}userStringCalc10", False, "read_attrs", 24],
        [ScalcoutRecord, f"{IOC}userStringCalc10", False, "configuration_attrs", 114],
        [ScalcoutRecord, f"{IOC}userStringCalc10", True, "read()", 3],
        [ScalcoutRecord, f"{IOC}userStringCalc10", True, "summary()", 250],

        [UserScalcoutDevice, IOC, False, "read_attrs", 250],
        [UserScalcoutDevice, IOC, False, "configuration_attrs", 1160],
        [UserScalcoutDevice, IOC, True, "read()", 30],
        [UserScalcoutDevice, IOC, True, "summary()", 2355],
    ]
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


def test_scalcout_reset():
    user = UserScalcoutDevice(IOC, name="user")
    user.wait_for_connection()
    user.enable.put("Enable")

    calc = user.scalcout10
    assert isinstance(calc, ScalcoutRecord)

    calc.reset()
    short_delay_for_EPICS_IOC_database_processing()
    assert calc.enable.get() in [1, "E"]

    # set some things
    calc.description.put("unit testing")
    v = 1.23456
    calc.channels.A.input_value.put(v)
    calc.channels.BB.input_value.put("testing")
    calc.calculation.put("A")
    calc.process_record.put(1)
    short_delay_for_EPICS_IOC_database_processing()
    assert calc.channels.A.input_value.get() == v
    assert calc.channels.BB.input_value.get() == "testing"
    assert calc.calculation.get() == "A"
    assert round(calc.calculated_value.get(), 5) == v
    assert calc.calculated_value_string.get() == str(v)

    calc.reset()
    short_delay_for_EPICS_IOC_database_processing()
    assert calc.description.get() == calc.prefix
    assert calc.calculation.get() == "0"
    assert calc.channels.A.input_value.get() == 0
    assert calc.channels.BB.input_value.get() == ""
    assert round(calc.calculated_value.get(), 5) == 0
    assert calc.calculated_value_string.get() == "0.00000"

    v1 = calc.calculated_value.get()
    short_delay_for_EPICS_IOC_database_processing(0.2)
    assert v1 == calc.calculated_value.get()



"""
PTC10 Programmable Temperature Controller
+++++++++++++++++++++++++++++++++++++++++

The PTC10 is a programmable temperature controller from SRS (Stanford Research
Systems).  The PTC10 is a modular system consisting of a base unit and provision
for addition of add-on boards.

A single, complete ``ophyd.Device`` subclass will not describe all variations
that could be installed. But the add-on boards each allow standardization. Each
installation must build a custom class that matches their hardware
configuration.  The APS USAXS instrument has created a `custom class
<https://github.com/APS-USAXS/ipython-usaxs/blob/master/profile_bluesky/startup/instrument/devices/ptc10_controller.py#L59-L159>`_
based on the `ophyd.PVPositioner` to use their `PTC10` as a temperature
*positioner*.

.. autosummary::

   ~PTC10AioChannel
   ~PTC10RtdChannel
   ~PTC10TcChannel
   ~PTC10PositionerMixin

:see: https://www.thinksrs.com/products/ptc10.html

EXAMPLE::

    from ophyd import PVPositioner

    class MyPTC10(PTC10PositionerMixin, PVPositioner):
        readback = Component(EpicsSignalRO, "2A:temperature", kind="hinted")
        setpoint = Component(EpicsSignalWithRBV, "5A:setPoint", kind="hinted")

        rtd = Component(PTC10RtdChannel, "3A:")
        pid = Component(PTC10AioChannel, "5A:")

    ptc10 = MyPTC10("IOC_PREFIX:ptc10:", name="ptc10")
    ptc10.report_dmov_changes.put(True)  # a diagnostic
    ptc10.tolerance.put(1.0)  # done when |readback-setpoint|<=tolerance

New in apstools 1.5.3.
"""

import logging

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import EpicsSignalWithRBV
from ophyd import Signal

logger = logging.getLogger(__name__)


class PTC10AioChannel(Device):
    """
    SRS PTC10 AIO module
    """

    voltage = Component(EpicsSignalRO, "voltage_RBV", kind="config")
    highlimit = Component(EpicsSignalWithRBV, "highLimit", kind="config")
    lowlimit = Component(EpicsSignalWithRBV, "lowLimit", kind="config")
    iotype = Component(EpicsSignalWithRBV, "ioType", kind="config", string=True)
    setpoint = Component(EpicsSignalWithRBV, "setPoint", kind="config")
    ramprate = Component(EpicsSignalWithRBV, "rampRate", kind="config")
    offswitch = Component(EpicsSignal, "off", kind="config")

    pidmode = Component(EpicsSignalWithRBV, "pid:mode", kind="config", string=True)
    P = Component(EpicsSignalWithRBV, "pid:P", kind="config")
    I = Component(EpicsSignalWithRBV, "pid:I", kind="config")
    D = Component(EpicsSignalWithRBV, "pid:D", kind="config")

    inputchoice = Component(EpicsSignalWithRBV, "pid:input", kind="config", string=True)
    tunelag = Component(EpicsSignalWithRBV, "tune:lag", kind="config")
    tunestep = Component(EpicsSignalWithRBV, "tune:step", kind="config")
    tunemode = Component(EpicsSignalWithRBV, "tune:mode", kind="config", string=True)
    tunetype = Component(EpicsSignalWithRBV, "tune:type", kind="config", string=True)


class PTC10RtdChannel(Device):
    """
    SRS PTC10 RTD module channel
    """

    temperature = Component(EpicsSignalRO, "temperature", kind="normal")
    units = Component(EpicsSignalRO, "units_RBV", kind="config", string=True)
    sensor = Component(EpicsSignalWithRBV, "sensor", kind="config", string=True)
    channelrange = Component(EpicsSignalWithRBV, "range", kind="config", string=True)
    current = Component(EpicsSignalWithRBV, "current", kind="config", string=True)
    power = Component(EpicsSignalWithRBV, "power", kind="config", string=True)


class PTC10TcChannel(Device):
    """
    SRS PTC10 Tc (thermocouple) module channel
    """

    temperature = Component(EpicsSignalRO, "temperature", kind="normal")


class PTC10PositionerMixin(Device):
    """
    Mixin so SRS PTC10 can be used as a (temperature) positioner.

    .. autosummary::

       ~cb_readback
       ~cb_setpoint
       ~inposition
       ~stop
    """

    done = Component(Signal, value=True, kind="omitted")
    done_value = True

    # for computation of soft `done` signal
    # default +/- 1 degree for "at temperature"
    tolerance = Component(Signal, value=1, kind="config")

    # For logging when temperature is reached after a move.
    report_dmov_changes = Component(Signal, value=True, kind="omitted")

    def cb_readback(self, *args, **kwargs):
        """
        Called when readback changes (EPICS CA monitor event).
        """
        diff = self.readback.get() - self.setpoint.get()
        dmov = abs(diff) <= self.tolerance.get()
        if self.report_dmov_changes.get() and dmov != self.done.get():
            logger.debug(f"{self.name} reached: {dmov}")
        self.done.put(dmov)

    def cb_setpoint(self, *args, **kwargs):
        """
        Called when setpoint changes (EPICS CA monitor event).

        When the setpoint is changed, force ``done=False``.  For any move,
        ``done`` MUST change to ``!= done_value``, then change back to
        ``done_value (True)``.  Without this response, a small move
        (within tolerance) will not return.  Next update of readback
        will compute ``self.done``.
        """
        self.done.put(not self.done_value)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readback.name = self.name

        # to compute the soft `done` signal
        self.readback.subscribe(self.cb_readback)
        self.setpoint.subscribe(self.cb_setpoint)

    @property
    def inposition(self):
        """
        Report (boolean) if positioner is done.
        """
        return self.done.get() == self.done_value

    def stop(self, *, success=False):
        """
        Hold the current readback when the stop() method is called and not done.
        """
        if not self.done.get():
            self.setpoint.put(self.position)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
EPICS PV Registry
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~findbyname
   ~findbypv
   ~PVRegistry
"""

import logging
import ophyd

from collections import defaultdict

from . import ipython_shell_namespace
from . import full_dotted_name


logger = logging.getLogger(__name__)


_findpv_registry = None


class PVRegistry:
    """
    Cross-reference EPICS PVs with ophyd EpicsSignalBase objects.
    """

    def __init__(self, ns=None):
        """
        Search ophyd objects for PV or ophyd names.

        The rebuild starts with the IPython console namespace
        (and defaults to the global namespace if the former
        cannot be obtained).

        PARAMETERS

        ns *dict* or `None`: namespace dictionary
        """
        self._pvdb = defaultdict(lambda: defaultdict(list))
        self._odb = {}
        self._device_name = None
        self._known_device_names = []
        g = ns or ipython_shell_namespace() or globals()

        # kickoff the registration process
        # fmt: off
        logger.debug(
            "Cross-referencing EPICS PVs with Python objects & ophyd symbols"
        )
        # fmt: on
        self._ophyd_epicsobject_walker(g)

    def _ophyd_epicsobject_walker(self, parent):
        """
        Walk through the parent object for ophyd Devices & EpicsSignals.

        This function is used to rebuild the ``self._pvdb`` object.
        """
        if isinstance(parent, dict):
            keys = parent.keys()
        else:
            keys = parent.component_names
            _nm_base = self._device_name
        for k in keys:
            if isinstance(parent, dict):
                _nm_base = []
                v = self._ref_dict(parent, k)
            else:
                v = self._ref_object_attribute(parent, k)
            if v is None:
                continue
            # print(k, type(v))
            if isinstance(v, ophyd.signal.EpicsSignalBase):
                try:
                    self._signal_processor(v)
                    self._odb[v.name] = ".".join(self._device_name + [k])
                except (KeyError, RuntimeError) as exc:
                    # fmt: off
                    logger.error(
                        "Exception while examining key '%s': (%s)", k, exc
                    )
                    # fmt: on
            elif isinstance(v, ophyd.Device):
                # print("Device", v.name)
                self._device_name = _nm_base + [k]
                if v.name not in self._known_device_names:
                    self._odb[v.name] = ".".join(self._device_name)
                    self._known_device_names.append(v.name)
                    self._ophyd_epicsobject_walker(v)

    def _ref_dict(self, parent, key):
        """Accessor used by ``_ophyd_epicsobject_walker()``"""
        return parent[key]

    def _ref_object_attribute(self, parent, key):
        """Accessor used by ``_ophyd_epicsobject_walker()``"""
        try:
            obj = getattr(parent, key, None)
            return obj
        except (KeyError, RuntimeError, TimeoutError) as exc:
            # fmt: off
            logger.error(
                "Exception while getting object '%s.%s': (%s)",
                ".".join(self._device_name), key, exc,
            )
            # fmt: on

    def _register_signal(self, signal, pv, mode):
        """Register a signal with the given mode."""
        fdn = full_dotted_name(signal)
        if fdn not in self._pvdb[pv][mode]:
            self._pvdb[pv][mode].append(fdn)

    def _signal_processor(self, signal):
        """Register a signal's read & write PVs."""
        self._register_signal(signal, signal._read_pv.pvname, "R")
        if hasattr(signal, "_write_pv"):
            self._register_signal(signal, signal._write_pv.pvname, "W")

    def search_by_mode(self, pvname, mode="R"):
        """Search for PV in specified mode."""
        if mode not in ["R", "W"]:
            raise ValueError(
                f"Incorrect mode given ({mode}." "  Must be either `R` or `W`."
            )
        return self._pvdb[pvname][mode]

    def search(self, pvname):
        """Search for PV in both read & write modes."""
        return dict(
            read=self.search_by_mode(pvname, "R"),
            write=self.search_by_mode(pvname, "W"),
        )

    def ophyd_search(self, oname):
        """Search for ophyd object by ophyd name."""
        return self._odb.get(oname)


def _get_pv_registry(force_rebuild, ns):
    """
    Check if need to build/rebuild the PV registry.

    PARAMETERS

    force_rebuild
        *bool* :
        If ``True``, rebuild the internal registry that maps
        EPICS PV names to ophyd objects.
    ns
        *dict* or `None` :
        Namespace dictionary of Python objects.

    """
    global _findpv_registry
    if _findpv_registry is None or force_rebuild:
        _findpv_registry = PVRegistry(ns=ns)
    return _findpv_registry


def findbyname(oname, force_rebuild=False, ns=None):
    """
    Find the ophyd (dotted name) object associated with the given ophyd name.

    PARAMETERS

    oname
        *str* :
        ophyd name to search
    force_rebuild
        *bool* :
        If ``True``, rebuild the internal registry that maps
        ophyd names to ophyd objects.
    ns
        *dict* or `None` :
        Namespace dictionary of Python objects.

    RETURNS

    str or ``None``:
        Name of the ophyd object.

    EXAMPLE::

        In [45]: findbyname("adsimdet_cam_acquire")
        Out[45]: 'adsimdet.cam.acquire'

    (new in apstools 1.5.0)
    """
    return _get_pv_registry(force_rebuild, ns).ophyd_search(oname)


def findbypv(pvname, force_rebuild=False, ns=None):
    """
    Find all ophyd objects associated with the given EPICS PV.

    PARAMETERS

    pvname
        *str* :
        EPICS PV name to search
    force_rebuild
        *bool* :
        If ``True``, rebuild the internal registry that maps
        EPICS PV names to ophyd objects.
    ns
        *dict* or `None` :
        Namespace dictionary of Python objects.

    RETURNS

    dict or ``None``:
        Dictionary of matching ophyd objects, keyed by how the
        PV is used by the ophyd signal.  The keys are
        ``read`` and ``write``.

    EXAMPLE::

        In [45]: findbypv("ad:cam1:Acquire")
        Out[45]: {'read': [], 'write': ['adsimdet.cam.acquire']}

        In [46]: findbypv("ad:cam1:Acquire_RBV")
        Out[46]: {'read': ['adsimdet.cam.acquire'], 'write': []}

    """
    return _get_pv_registry(force_rebuild, ns).search(pvname)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Base Class for File Writer Callbacks
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~FileWriterCallbackBase
"""


import datetime
import logging
import os
import pyRestTable


logger = logging.getLogger(__name__)


class FileWriterCallbackBase:
    """
    Base class for filewriter callbacks.

    .. index:: Bluesky Callback; FileWriterCallbackBase

    New with apstools release 1.3.0.

    Applications should subclass and rewrite the ``writer()`` method.

    The local buffers are cleared when a start document is received.
    Content is collected here from each document until the stop document.
    The content is written once the stop document is received.


    User Interface methods

    .. autosummary::

       ~receiver

    Internal methods

    .. autosummary::

       ~clear
       ~make_file_name
       ~writer

    Document Handler methods

    .. autosummary::

       ~bulk_events
       ~datum
       ~descriptor
       ~event
       ~resource
       ~start
       ~stop
    """

    file_extension = "dat"
    file_name = None
    file_path = None

    # convention: methods written in alphabetical order

    def __init__(self, *args, **kwargs):
        """Initialize: clear and reset."""
        self.clear()
        self.xref = dict(
            bulk_events=self.bulk_events,
            datum=self.datum,
            descriptor=self.descriptor,
            event=self.event,
            resource=self.resource,
            start=self.start,
            stop=self.stop,
        )

    def receiver(self, key, doc):
        """
        bluesky callback (handles a stream of documents)

        .. index:: Bluesky Callback; FileWriterCallbackBase.receiver
        """
        handler = self.xref.get(key)
        if handler is None:
            logger.error("unexpected key %s" % key)
        else:
            handler(doc)

        # - - - - - - - - - - - - - - -

    def clear(self):
        """
        delete any saved data from the cache and reinitialize
        """
        self.acquisitions = {}
        self.detectors = []
        self.exit_status = None
        self.externals = {}
        self.metadata = {}
        self.plan_name = None
        self.positioners = []
        self.scanning = False
        self.scan_id = None
        self.streams = {}
        self.start_time = None
        self.stop_reason = None
        self.stop_time = None
        self.uid = None

    def make_file_name(self):
        """
        generate a file name to be used as default

        default format: {ymd}-{hms}-S{scan_id}-{short_uid}.{ext}
        where the time (the run start time):

        * ymd = {year:4d}{month:02d}{day:02d}
        * hms = {hour:02d}{minute:02d}{second:02d}

        override in subclass to change
        """
        start_time = datetime.datetime.fromtimestamp(self.start_time)
        fname = start_time.strftime("%Y%m%d-%H%M%S")
        fname += f"-S{self.scan_id:05d}"
        fname += f"-{self.uid[:7]}.{self.file_extension}"
        path = os.path.abspath(self.file_path or os.getcwd())
        return os.path.join(path, fname)

    def writer(self):
        """
        print summary of run as diagnostic

        override this method in subclass to write a file
        """
        # logger.debug("acquisitions: %s", yaml.dump(self.acquisitions))

        fname = self.file_name or self.make_file_name()
        print("print to console")
        print(f"suggested file name: {fname}")  # lgtm [py/clear-text-logging-sensitive-data]

        tbl = pyRestTable.Table()
        tbl.labels = "key value".split()
        keys = "plan_name scan_id exit_status start_time stop_reason stop_time uid".split()
        for k in sorted(keys):
            tbl.addRow((k, getattr(self, k)))
        print(tbl)

        def trim(value, length=60):
            text = str(value)
            if len(text) > length:
                text = text[: length - 3] + "..."
            return text

        tbl = pyRestTable.Table()
        tbl.labels = "metadata_key value".split()
        for k, v in sorted(self.metadata.items()):
            tbl.addRow((k, trim(v)))
        print(tbl)

        tbl = pyRestTable.Table()
        tbl.labels = "stream num_vars num_values".split()
        for k, v in sorted(self.streams.items()):
            if len(v) != 1:
                print(
                    f"expecting only one descriptor in stream {k},"
                    f" found {len(v)}"
                )
            else:
                data = self.acquisitions[v[0]]["data"]
                num_vars = len(data)
                # get the key (symbol) of first data object
                symbol = list(data.keys())[0]
                if num_vars == 0:
                    num_values = 0
                else:
                    num_values = len(data[symbol]["data"])
                tbl.addRow((k, num_vars, num_values))
        print(tbl)

        tbl = pyRestTable.Table()
        tbl.labels = "external_key type value".split()
        for k, v in self.externals.items():
            tbl.addRow((k, v["_document_type_"], trim(v)))
        print(tbl)

        print(f"elapsed scan time: {self.stop_time-self.start_time:.3f}s")

    # - - - - - - - - - - - - - - -

    def bulk_events(self, doc):
        """Deprecated. Use EventPage instead."""
        if not self.scanning:
            return
        logger.info("not handled yet")
        logger.info(doc)
        logger.info("-" * 40)

    def datum(self, doc):
        """
        Like an event, but for data recorded outside of bluesky.

        Example::

            Datum
            =====
            datum_id        : 621caa0f-70f1-4e3d-8718-b5123d434502/0
            datum_kwargs    :
            HDF5_file_name  : /mnt/usaxscontrol/USAXS_data/2020-06/06_10_Minjee_waxs/AGIX3N1_0699.hdf
            point_number    : 0
            resource        : 621caa0f-70f1-4e3d-8718-b5123d434502
        """
        if not self.scanning:
            return
        # stash the whole thing (sort this out in the writer)
        ext = self.externals[doc["datum_id"]] = dict(doc)
        ext["_document_type_"] = "datum"

    def descriptor(self, doc):
        """
        description of the data stream to be acquired
        """
        if not self.scanning:
            return
        stream = doc["name"]
        uid = doc["uid"]

        if stream not in self.streams:
            self.streams[stream] = []
        self.streams[stream].append(uid)

        if uid not in self.acquisitions:
            self.acquisitions[uid] = dict(stream=stream, data={})
        data = self.acquisitions[uid]["data"]
        for k, entry in doc["data_keys"].items():
            # logger.debug("entry %s: %s", k, entry)
            dd = data[k] = {}
            dd["source"] = entry.get("source", "local")
            dd["dtype"] = entry.get("dtype", "")
            dd["shape"] = entry.get("shape", [])
            dd["units"] = entry.get("units", "")
            dd["lower_ctrl_limit"] = entry.get("lower_ctrl_limit", "")
            dd["upper_ctrl_limit"] = entry.get("upper_ctrl_limit", "")
            dd["precision"] = entry.get("precision", 0)
            dd["object_name"] = entry.get("object_name", k)
            dd["data"] = []  # entry data goes here
            dd["time"] = []  # entry time stamps here
            dd["external"] = entry.get("external") is not None
            # logger.debug("dd %s: %s", k, data[k])

    def event(self, doc):
        """
        a single "row" of data
        """
        if not self.scanning:
            return
        # uid = doc["uid"]
        descriptor_uid = doc["descriptor"]
        # seq_num = doc["seq_num"]

        # gather the data by streams
        descriptor = self.acquisitions.get(descriptor_uid)
        if descriptor is not None:
            for k, v in doc["data"].items():
                data = descriptor["data"].get(k)
                if data is None:
                    print(
                        f"entry key {k} not found in"
                        f" descriptor of {descriptor['stream']}"
                    )
                else:
                    data["data"].append(v)
                    data["time"].append(doc["timestamps"][k])

    def resource(self, doc):
        """
        like a descriptor, but for data recorded outside of bluesky
        """
        if not self.scanning:
            return
        # stash the whole thing (sort this out in the writer)
        ext = self.externals[doc["uid"]] = dict(doc)
        ext["_document_type_"] = "resource"

    def start(self, doc):
        """
        beginning of a run, clear cache and collect metadata
        """
        self.clear()
        self.plan_name = doc["plan_name"]
        self.scanning = True
        self.scan_id = doc["scan_id"] or 0
        self.start_time = doc["time"]
        self.uid = doc["uid"]
        self.detectors = doc.get("detectors", [])
        self.positioners = (
            doc.get("positioners") or doc.get("motors") or []
        )

        # gather the metadata
        for k, v in doc.items():
            if k in "scan_id time uid".split():
                continue
            self.metadata[k] = v

    def stop(self, doc):
        """
        end of the run, end collection and initiate the ``writer()`` method
        """
        if not self.scanning:
            return
        self.exit_status = doc["exit_status"]
        self.stop_reason = doc.get("reason", "not available")
        self.stop_time = doc["time"]
        self.scanning = False

        self.writer()

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Directory of the known plans
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~listplans
"""

import inspect
import logging
from ophyd.ophydobj import OphydObject
import pandas as pd

logger = logging.getLogger(__name__)


def listplans(base=None, trunc=40):
    """
    List all plans.  (Actually, lists all generator functions).

    NOTE: Can only detect generator functions.
    Bluesky plans are generator functions that generate
    ``bluesky.Msg`` objects.  There is a PR to define
    a decorator that identifies a generator function
    as a bluesky plan.

    PARAMETERS

    base
        *object* or *None* :
        Object that contains plan methods (if ``None``, use global namespace.)
        (default: ``None``)
    trunc
        *int* :
        Truncate long docstrings to no more than ``trunc`` characters.
        (default: 40)
    """
    if base is None:
        try:
            from IPython import get_ipython

            base = get_ipython().user_global_ns
            # logger.debug("IPython user global namespace: %s", base.keys())
        except (ModuleNotFoundError, AttributeError):
            base = globals()
            # logger.debug("globals() namespace: %s", base.keys())
        _gg = {k: v for k, v in base.items() if not k.startswith("_")}
    else:
        # fmt: off
        _gg = {
            k: getattr(base, k)
            for k in dir(base)
            if not k.startswith("_")  # not interested in these
            if k not in ("signal_names", )  # names to avoid
        }
        # fmt: on

    if inspect.ismodule(base):
        prefix = base.__name__ + "."
    elif isinstance(base, OphydObject):
        prefix = base.name + "."
    else:
        prefix = ""

    dd = dict(plan=[], doc=[])

    for key, obj in _gg.items():
        if inspect.isgeneratorfunction(obj):  # TODO: bluesky.isplan(obj)
            doc = (obj.__doc__ or "").split("\n")[0]
            if len(doc.strip()) == 0:
                doc = "---"
            if len(doc) > trunc and trunc > 4:
                doc = doc[: trunc - 4] + " ..."
            dd["plan"].append(f"{prefix}{key}")
            dd["doc"].append(doc)
    return pd.DataFrame(dd)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Unit tests for :mod:`~apstools._utils.device_info`.
"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsMotor
from ophyd import Signal
from ophyd.signal import EpicsSignalBase
import pandas as pd
import pytest

from ...devices import SwaitRecord
from ...tests import MASTER_TIMEOUT
from ..device_info import _list_epics_signals
from ..device_info import listdevice


# set default timeout for all EpicsSignal connections & communications
try:
    EpicsSignalBase.set_defaults(
        auto_monitor=True,
        timeout=MASTER_TIMEOUT,
        write_timeout=MASTER_TIMEOUT,
        connection_timeout=MASTER_TIMEOUT,
    )
except RuntimeError:
    pass  # ignore if some EPICS object already created


IOC = "gp:"  # for testing with an EPICS IOC


class MySignals(Device):
    allowed = Component(Signal, value=True, kind="omitted")
    background = Component(Signal, value=True, kind="normal")
    tolerance = Component(Signal, value=1, kind="config")
    visible = Component(Signal, value=True, kind="hinted")


class MyDevice(Device):
    signals = Component(MySignals)
    calc5 = Component(SwaitRecord, "5")
    calc6 = Component(SwaitRecord, "6")


calcs = MyDevice(f"{IOC}userCalc", name="calcs")
motor = EpicsMotor(f"{IOC}m1", name="motor")
signal = Signal(name="signal", value=True)

calcs.wait_for_connection()
motor.wait_for_connection()


def test_calcs():
    assert calcs.connected
    assert calcs is not None
    assert isinstance(calcs, Device)


@pytest.mark.parametrize(
    "obj, length",
    [
        (calcs, 4),
        (calcs.calc5.description, 1),
        (signal, 1),
        (motor, 2),
        (motor.user_setpoint, 1),
    ],
)
def test_listdevice(obj, length):
    result = listdevice(obj, scope="read")
    assert isinstance(result, pd.DataFrame)
    assert len(result) == length
    if length > 0:
        assert len(result.columns) == 3
        expected = ["data name", "value", "timestamp"]
        for r in result.columns:
            assert r in expected


@pytest.mark.parametrize(
    "obj, length, ref",
    [
        (calcs, 126, EpicsSignalBase),
        (calcs.calc5.description, 1, EpicsSignalBase),
        (signal, None, None),
        (motor, 19, EpicsSignalBase),
        (motor.user_setpoint, 1, EpicsSignalBase),
    ],
)
def test__list_epics_signals(obj, length, ref):
    result = _list_epics_signals(obj)
    if length is None:
        assert result is None
    else:
        assert isinstance(result, list)
        assert len(result) == length
        for item in result:
            assert isinstance(item, ref)


@pytest.mark.parametrize(
    "scope, row, column, value",
    [
        (None, 0, "data name", "motor"),
        (None, 0, "value", 0.0),
        ("epics", 0, "data name", "motor"),
        ("epics", 0, "value", 0.0),
        ("epics", 7, "data name", "motor_acceleration"),
        ("epics", 7, "value", 0.2),
        ("epics", 8, "data name", "motor_motor_egu"),
        ("epics", 8, "value", "degrees"),
        ("full", 0, "data name", "motor"),
        ("full", 0, "value", 0.0),
        ("full", 12, "data name", "motor_low_limit_switch"),
        ("full", 12, "value", 0),
        ("read", 0, "data name", "motor"),
        ("read", 0, "value", 0.0),
    ],
)  # scope: full epics read
def test_spotchecks(scope, row, column, value):
    assert calcs.connected
    if round(motor.position, 2) != 0:
        motor.move(0)
    result = listdevice(motor, scope=scope)
    assert isinstance(result, pd.DataFrame)
    assert column in result
    assert row in result[column]
    assert result[column][row] == value


@pytest.mark.parametrize(
    "device, scope, ancient, length",
    [
        (calcs, "epics", False, 0),
        (calcs, "epics", True, 126),
        (calcs, "full", False, 4),
        (calcs, "full", True, 130),
        (calcs, "read", False, 2),
        (calcs, "read", True, 4),
        (calcs, None, False, 4),
        (calcs, None, True, 130),
    ],
)
def test_listdevice_filters(device, scope, ancient, length):
    result = listdevice(device, scope, show_ancient=ancient)
    assert len(result) == length


@pytest.mark.parametrize(
    "device, scope, cnames",
    [
        (calcs, "full", [
            "calcs.signals.allowed",
            "calcs.signals.background",
            "calcs.signals.tolerance",
            "calcs.signals.visible",
        ]),
        # This is the case that produced issue 640
        (calcs.signals, "full", [
            "calcs.signals.allowed",
            "calcs.signals.background",
            "calcs.signals.tolerance",
            "calcs.signals.visible",
        ]),
    ],
)
def test_listdevice_cname(device, scope, cnames):
    result = listdevice(device, scope, show_ancient=False, cname=True)
    assert result["name"].tolist() == cnames



"""
Ophyd support for the EPICS transform record


Public Structures

.. autosummary::

    ~UserTransformN
    ~UserTransformsDevice
    ~TransformRecord
"""

from collections import OrderedDict
from ophyd import Device
from ophyd import Component as Cpt
from ophyd import DynamicDeviceComponent as DDC
from ophyd import FormattedComponent as FC
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO

from ._common import EpicsRecordDeviceCommonAll
from ._common import EpicsSynAppsRecordEnableMixin
from .. import utils as APS_utils


CHANNEL_LETTERS_LIST = "A B C D E F G H I J K L M N O P".split()


class transformRecordChannel(Device):
    """
    channel of a synApps transform record: A-P

    .. index:: Ophyd Device; synApps transformRecordChannel

    .. autosummary::

        ~reset
    """

    current_value = FC(EpicsSignal, "{self.prefix}.{self._ch_letter}", kind="hinted")
    last_value = FC(EpicsSignalRO, "{self.prefix}.L{self._ch_letter}", kind="config")
    input_pv = FC(EpicsSignal, "{self.prefix}.INP{self._ch_letter}", kind="config")
    input_pv_valid = FC(
        EpicsSignalRO, "{self.prefix}.I{self._ch_letter}V", kind="config"
    )
    expression_invalid = FC(
        EpicsSignalRO, "{self.prefix}.C{self._ch_letter}V", kind="config"
    )
    comment = FC(EpicsSignal, "{self.prefix}.CMT{self._ch_letter}", kind="config")
    expression = FC(EpicsSignal, "{self.prefix}.CLC{self._ch_letter}", kind="config")
    output_pv = FC(EpicsSignal, "{self.prefix}.OUT{self._ch_letter}", kind="config")
    output_pv_valid = FC(
        EpicsSignalRO, "{self.prefix}.O{self._ch_letter}V", kind="config"
    )

    read_attrs = ["current_value"]

    def __init__(self, prefix, letter, **kwargs):
        self._ch_letter = letter
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.comment.put(self._ch_letter.lower())
        self.input_pv.put("")
        self.expression.put("")
        self.current_value.put(0)
        self.output_pv.put("")


def _channels(channel_list):
    defn = OrderedDict()
    for chan in channel_list:
        defn[chan] = (transformRecordChannel, "", {"letter": chan})
    return defn


class TransformRecord(EpicsRecordDeviceCommonAll):
    """
    EPICS transform record support in ophyd

    .. index:: Ophyd Device; synApps TransformRecord

    .. autosummary::

        ~reset

    :see: https://htmlpreview.github.io/?https://raw.githubusercontent.com/epics-modules/calc/R3-6-1/documentation/TransformRecord.html#Fields
    """

    units = Cpt(EpicsSignal, ".EGU", kind="config")
    precision = Cpt(EpicsSignal, ".PREC", kind="config")
    version = Cpt(EpicsSignalRO, ".VERS", kind="config")

    calc_option = Cpt(EpicsSignal, ".COPT", kind="config")
    invalid_link_action = Cpt(EpicsSignalRO, ".IVLA", kind="config")
    input_bitmap = Cpt(EpicsSignalRO, ".MAP", kind="config")

    read_attrs = APS_utils.itemizer("channels.%s", CHANNEL_LETTERS_LIST)
    hints = {"fields": read_attrs}

    channels = DDC(_channels(CHANNEL_LETTERS_LIST))

    def reset(self):
        """set all fields to default values"""
        self.scanning_rate.put("Passive")
        self.description.put(self.description.pvname.split(".")[0])
        self.units.put("")
        self.calc_option.put(0)
        self.precision.put("3")
        self.forward_link.put("")
        for letter in self.channels.read_attrs:
            channel = getattr(self.channels, letter)
            if isinstance(channel, transformRecordChannel):
                channel.reset()
        self.hints["fields"] = ["channels.%s" % c for c in CHANNEL_LETTERS_LIST]
        self.read_attrs = ["channels.%s" % c for c in CHANNEL_LETTERS_LIST]


class UserTransformN(EpicsSynAppsRecordEnableMixin, TransformRecord):
    """Single instance of the userTranN database."""


class UserTransformsDevice(Device):
    """
    EPICS synApps XXX IOC setup of userTransforms: ``$(P):userTran$(N)``

    .. index:: Ophyd Device; synApps UserTransformsDevice
    """

    enable = Cpt(EpicsSignal, "userTranEnable", kind="config")
    transform1 = Cpt(UserTransformN, "userTran1")
    transform2 = Cpt(UserTransformN, "userTran2")
    transform3 = Cpt(UserTransformN, "userTran3")
    transform4 = Cpt(UserTransformN, "userTran4")
    transform5 = Cpt(UserTransformN, "userTran5")
    transform6 = Cpt(UserTransformN, "userTran6")
    transform7 = Cpt(UserTransformN, "userTran7")
    transform8 = Cpt(UserTransformN, "userTran8")
    transform9 = Cpt(UserTransformN, "userTran9")
    transform10 = Cpt(UserTransformN, "userTran10")

    def reset(self):  # lgtm [py/similar-function]
        """set all fields to default values"""
        self.transform1.reset()
        self.transform2.reset()
        self.transform3.reset()
        self.transform4.reset()
        self.transform5.reset()
        self.transform6.reset()
        self.transform7.reset()
        self.transform8.reset()
        self.transform9.reset()
        self.transform10.reset()
        self.read_attrs = ["transform%d" % (c + 1) for c in range(10)]
        self.read_attrs.insert(0, "enable")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
simple unit tests for this package
"""

import databroker
import numpy as np
import ophyd.sim
import pytest
import time
import uuid

from ... import __version__ as APS__version__
from ... import utils
from .._core import MAX_EPICS_STRINGOUT_LENGTH

CATALOG = "usaxs_test"
COUNT = "555a604"  # <-- uid,  scan_id: 2
RE = None
DEFAULT_CATALOG_ID = uuid.uuid4()  # a unique ID for testing purposes


@pytest.fixture(scope="function")
def cat():
    cat = databroker.catalog[CATALOG]
    return cat


def test_utils_cleanupText():
    original = "1. Some text to cleanup #25"
    received = utils.cleanupText(original)
    expected = "1__Some_text_to_cleanup__25"
    assert received == expected


def test_utils_dictionary_table():
    md = {
        "login_id": "jemian:wow.aps.anl.gov",
        "beamline_id": "developer",
        "proposal_id": None,
        "pid": 19072,
        "scan_id": 10,
        "version": {
            "bluesky": "1.5.2",
            "ophyd": "1.3.3",
            "apstools": "1.1.5",
            "epics": "3.3.3",
        },
    }
    table = utils.dictionary_table(md)
    received = str(table).strip()
    expected = (
        "=========== =============================================================================\n"
        "key         value                                                                        \n"
        "=========== =============================================================================\n"
        "beamline_id developer                                                                    \n"
        "login_id    jemian:wow.aps.anl.gov                                                       \n"
        "pid         19072                                                                        \n"
        "proposal_id None                                                                         \n"
        "scan_id     10                                                                           \n"
        "version     {'bluesky': '1.5.2', 'ophyd': '1.3.3', 'apstools': '1.1.5', 'epics': '3.3.3'}\n"
        "=========== ============================================================================="
    )
    assert received == expected


def test_utils_itemizer():
    items = [1.0, 1.1, 1.01, 1.001, 1.0001, 1.00001]
    received = utils.itemizer("%.2f", items)
    expected = ["1.00", "1.10", "1.01", "1.00", "1.00", "1.00"]
    assert received == expected


def test_utils_print_RE_md(capsys):
    global RE
    md = {}
    md["purpose"] = "testing"
    md["versions"] = dict(apstools=APS__version__)
    md["something"] = "else"

    utils.print_RE_md(md)
    out, err = capsys.readouterr()
    received = out.splitlines()

    expected = [
        "RunEngine metadata dictionary:",
        "========= ===============================",
        "key       value                          ",
        "========= ===============================",
        "purpose   testing                        ",
        "something else                           ",
        "versions  ======== ======================",
        "          key      value                 ",
        "          ======== ======================",
        f"          apstools {APS__version__}",
        "          ======== ======================",
        "========= ===============================",
        "",
    ]
    assert len(received) == len(expected)
    assert received[4].strip() == expected[4].strip()
    assert received[5].strip() == expected[5].strip()
    assert received[9].strip() == expected[9].strip()


def test_utils_pairwise():
    items = [1.0, 1.1, 1.01, 1.001, 1.0001, 1.00001, 2]
    received = list(utils.pairwise(items))
    expected = [(1.0, 1.1), (1.01, 1.001), (1.0001, 1.00001)]
    assert received == expected


@pytest.mark.parametrize(
    "given, expected",
    [
        ["simple", "simple"],
        ["sim_ple", "sim_ple"],
        ["hy-phen", "hy_phen"],
        ["white space", "white_space"],
        ["1world", "_1world"],
        ["end9", "end9"],
        ["$tree", "_tree"],
        ["#!bang", "__bang"],
        ["0 is not a good name", "_0_is_not_a_good_name"],
        ["! is even worse!", "__is_even_worse_"],
        ["double.dotted.name", "double_dotted_name"],
    ],
)
def test_utils_safe_ophyd_name(given, expected):
    received = utils.safe_ophyd_name(given)
    assert received == expected


def test_utils_split_quoted_line():
    source = 'FlyScan 5   2   0   "empty container"'
    received = utils.split_quoted_line(source)
    assert len(received) == 5
    expected = ["FlyScan", "5", "2", "0", "empty container"]
    assert received == expected


def test_utils_trim_string_for_EPICS():
    source = "0123456789"
    assert len(source) < MAX_EPICS_STRINGOUT_LENGTH
    received = utils.trim_string_for_EPICS(source)
    assert len(source) == len(received)
    expected = source
    assert received == expected

    source = "0123456789" * 10
    assert len(source) > MAX_EPICS_STRINGOUT_LENGTH
    received = utils.trim_string_for_EPICS(source)
    assert len(source) > len(received)
    expected = source[: MAX_EPICS_STRINGOUT_LENGTH - 1]
    assert received == expected


def test_utils_listobjects():
    sims = ophyd.sim.hw().__dict__
    wont_show = (
        "flyer1",
        "flyer2",
        "new_trivial_flyer",
        "trivial_flyer",
    )
    num = len(sims) - len(wont_show)
    kk = sorted(sims.keys())

    table = utils.listobjects(symbols=sims, printing=False)
    assert 4 == len(table.labels)
    rr = [r[0] for r in table.rows]
    for k in kk:
        if k not in wont_show:
            assert k in rr
    assert num == len(table.rows)


def test_utils_unix():
    cmd = 'echo "hello"'
    out, err = utils.unix(cmd)
    assert out == b"hello\n"
    assert err == b""

    cmd = "sleep 0.8 | echo hello"
    t0 = time.time()
    out, err = utils.unix(cmd)
    dt = time.time() - t0
    assert dt >= 0.8
    assert out == b"hello\n"
    assert err == b""


def test_utils_with_database_listruns(cat):
    assert len(list(cat.v1[COUNT].documents())[:1]) == 1
    df = utils.listruns(cat=cat, printing=False, num=10)
    assert df is not None
    assert len(df.columns) == 4

    assert "time" in df
    assert df.columns[1] == "time"
    ts = df["time"]
    assert len(ts) == 10
    assert (ts == np.sort(ts)[::-1]).all()


def test_utils_with_database_replay(cat):
    replies = []

    def cb1(key, doc):
        replies.append((key, len(doc)))

    utils.replay(cat.v1[COUNT], callback=cb1)
    assert len(replies) > 0
    keys = set([v[0] for v in replies])
    # doc_types = "start stop event descriptor datum resource".split()
    doc_types = "start stop event descriptor".split()
    for item in doc_types:
        assert item in keys

    def cb2(key, doc):
        if key == "start":
            replies.append(doc["time"])

    replies = []
    utils.replay(cat.v1[-1], callback=cb2)
    assert len(replies) == 1  # last run

    replies = []
    utils.replay(cat.v1[-3:], callback=cb2)
    assert len(replies) == 3  # last 3 runs

    replies = []
    utils.replay(cat.v1[COUNT], callback=cb2)
    previous = 0
    for i, v in enumerate(replies):
        assert v - previous > 0
        previous = v

    replies = []
    utils.replay(cat.v1[-1], callback=cb2, sort=False)
    previous = replies[0] + 1
    for i, v in enumerate(replies):
        assert v - previous < 0
        previous = v


@pytest.mark.parametrize(
    "run_type, ref, scan_ids",
    [
        ["header", -1, 1],
        ["run", -1, 1],
        ["run", 1, 1],
        ["run", "e5d2cbdc", 1],
        ["run", -2, 110],
        ["run", 110, 110],
        ["run", "19965989-0a2a-44aa-aa06-c1248754e651", 110],
        ["header", [-1, -2], [1, 110]],
        ["run", -3, 109],
        ["run", [1, 110, 109], [1, 110, 109]],
    ],
)
def test_replay(run_type, ref, scan_ids, cat):
    if run_type == "header":
        local_cat = cat.v1
    else:
        local_cat = cat.v2
    if isinstance(ref, (list, tuple)):
        n_items = len(ref)
        local_ref = [local_cat[r] for r in ref]
        if run_type == "header":
            local_scan_ids = [r.start["scan_id"] for r in local_ref]
        else:
            local_scan_ids = [r.metadata["start"]["scan_id"] for r in local_ref]
    else:
        n_items = 1
        local_ref = local_cat[ref]
        if run_type == "header":
            local_scan_ids = local_ref.start["scan_id"]
        else:
            local_scan_ids = local_ref.metadata["start"]["scan_id"]
    assert local_scan_ids == scan_ids

    replies = []

    def cb2(key, doc):
        if key == "start":
            replies.append(doc["time"])

    utils.replay(local_ref, cb2)
    assert len(replies) == n_items


@pytest.mark.parametrize(
    "scan_id, stream, total_keys, key, v1, m3, m_default, m_strict, m_lower",
    [
        (103, "primary", 7, "PD_USAXS", False, 1, 1, 1, 0),
        (103, None, 7, "PD_USAXS", False, 1, 1, 1, 0),
        (2, "primary", 5, "scaler0", False, 2, 2, 2, 2),
        (2, None, 5, "scaler0", False, 2, 2, 2, 2),
        (103, "primary", 8, "PD_USAXS", True, 1, 1, 1, 0),
    ],
)
def test_utils_listRunKeys(
    scan_id, stream, total_keys, key, v1, m3, m_default, m_strict, m_lower, cat
):
    assert (
        len(utils.listRunKeys(scan_id, db=cat, stream=stream, use_v1=v1)) == total_keys
    )

    result = utils.listRunKeys(
        scan_id, key_fragment=key[0:3], db=cat, stream=stream, use_v1=v1
    )
    assert len(result) == m3

    result = utils.listRunKeys(
        scan_id, key_fragment=key, db=cat, stream=stream, use_v1=v1
    )
    assert len(result) == m_default

    result = utils.listRunKeys(
        scan_id, key_fragment=key, db=cat, stream=stream, strict=True, use_v1=v1
    )
    assert len(result) == m_strict

    result = utils.listRunKeys(
        scan_id, key_fragment=key.lower(), db=cat, stream=stream, strict=True, use_v1=v1
    )
    assert len(result) == m_lower


# fmt: off
@pytest.mark.parametrize(
    "scan_id, stream",
    [
        (110, None),
        (110, "primary"),
        (103, "mca"),
    ],
)
def test_utils_listRunKeys_no_such_stream(scan_id, stream, cat):
    with pytest.raises(AttributeError) as exc:
        utils.listRunKeys(scan_id, db=cat, stream=stream)
    assert str(exc.value).startswith("No such stream ")
# fmt: on


@pytest.mark.parametrize(
    "scan_id, stream, nkeys, v1",
    [
        (2, None, 5, False),
        (2, "primary", 5, False),
        (2, "primary", 6, True),
        (103, "primary", 7, False),
        # (103, "baseline", 268, False),  # very slow test!
        (103, "baseline", 269, True),  # faster test!
        (103, "baseline", 269, None),  # faster test!
    ],
)
def test_utils_getRunData(scan_id, stream, nkeys, v1, cat):
    table = utils.getRunData(scan_id, db=cat, stream=stream, use_v1=v1)
    assert table is not None
    assert len(table.keys()) == nkeys


@pytest.mark.parametrize(
    "scan_id, stream, key, idx, v1, expected, prec",
    [
        # (2, "baseline", "undulator_downstream_version", None, False, "4.21", 0),  # VERY slow
        (2, "baseline", "undulator_downstream_version", None, True, "4.21", 0),
        (2, "primary", "I0_USAXS", -1, False, 3729, 0),
        (2, "primary", "I0_USAXS", "-1", False, 3729, 0),
        (
            2,
            "primary",
            "I0_USAXS",
            "all",
            False,
            [
                3729.0,
            ],
            0,
        ),
        (2, "primary", "I0_USAXS", None, False, 3729, 0),
        (
            2,
            None,
            "I0_USAXS",
            "all",
            False,
            [
                3729.0,
            ],
            0,
        ),
        # (103, "baseline", "undulator_downstream_version", None, False, "4.21", 0),  # VERY slow
        (103, "baseline", "undulator_downstream_version", None, True, "4.21", 0),
        (103, "primary", "a_stage_r", -1, False, 8.88197, 5),
        (103, "primary", "a_stage_r", -1, True, 8.88197, 5),
        (103, "primary", "a_stage_r", "mean", False, 8.88397, 5),
        (103, "primary", "a_stage_r", 0, False, 8.88597, 5),
        (103, "primary", "a_stage_r", None, False, 8.88197, 5),
        # (110, "baseline", "terms_SAXS_UsaxsSaxsMode", None, False, "blank", 0),  # VERY slow
        # (110, "baseline", "user_data_sample_thickness", None, False, 0.0, 1),  # VERY slow
        # (110, "baseline", "user_data_scan_macro", None, False, "FlyScan", 0),  # VERY slow
        (110, "baseline", "terms_SAXS_UsaxsSaxsMode", None, True, "blank", 0),
        (110, "baseline", "user_data_sample_thickness", None, True, 0.0, 1),
        (110, "baseline", "user_data_scan_macro", None, True, "FlyScan", 0),
    ],
)
def test_utils_getRunDataValue(scan_id, stream, key, idx, v1, expected, prec, cat):
    value = utils.getRunDataValue(
        scan_id, key, db=cat, stream=stream, idx=idx, use_v1=v1
    )
    if isinstance(value, str):
        assert value == expected
    elif isinstance(value, float):
        assert round(value, prec) == expected
    elif isinstance(value, np.ndarray):
        assert len(value) == len(expected)
        for v, e in zip(value, expected):
            assert round(v, prec) == e


# # fmt: off
# @pytest.mark.parametrize(
#     "scan_id, stream, key, idx, expected, prec",
#     [
#         (110, "mca", "struck_mca3_elapsed_real_time", None, 93.72, 2),
#     ],
# )
# def test_utils_getRunDataValue_MemoryError(
#     scan_id, stream, key, idx, expected, prec, cat
# ):
#     with pytest.raises(np.core._exceptions._ArrayMemoryError) as exc:
#         utils.getRunDataValue(
#             scan_id, key, db=cat, stream=stream, idx=idx
#         )
#     assert str(exc.value).startswith("Unable to allocate ")
# # fmt: on


# fmt: off
@pytest.mark.parametrize(
    "query, n",
    [
        ({"since": "2019-05-01"}, 8),
        ({"since": "2000-05-01"}, 10),
        ({}, 10),
        (None, 10),
        ({"since": None}, 10),
        ({"until": "2019-05-01"}, 2),
        ({"scan_id": 2}, 2),
        ({"scan_id": 2, "plan_name": "count"}, 1),
        ({"scan_id": 110}, 1),
    ],
)
def test_utils_db_query(
    query, n, cat
):
    sub_cat = utils.db_query(cat, query)
    assert len(sub_cat) == n
# fmt: on


# fmt: off
@pytest.mark.parametrize(
    "scan_id, key, db, stream, query, v1, nrows, ncols",
    [
        (COUNT, None, DEFAULT_CATALOG_ID, None, None, None, 6, 1),
        (COUNT, None, DEFAULT_CATALOG_ID, "baseline", None, None, 269, 2),
        (COUNT, "will not be found", DEFAULT_CATALOG_ID, "baseline", None, None, 1, 2),
        (COUNT, "aps", DEFAULT_CATALOG_ID, None, None, None, 1, 1),
        (COUNT, None, DEFAULT_CATALOG_ID, "primary", None, None, 6, 1),
        (COUNT, "aps", DEFAULT_CATALOG_ID, "primary", None, None, 1, 1),
    ],
)
def test_utils_getStreamValues(scan_id, key, db, stream, cat, query, v1, nrows, ncols):
    """These tests should not raise exceptions."""
    if db == DEFAULT_CATALOG_ID:
        db = cat
    table = utils.getStreamValues(
        scan_id, db=db, key_fragment=key, stream=stream, query=query, use_v1=v1
    )
    assert table is not None
    assert table.shape == (nrows, ncols)
# fmt: on


# fmt: off
@pytest.mark.parametrize(
    "scan_id, key, db, stream, query, v1, error, first_words",
    [
        (COUNT, None, DEFAULT_CATALOG_ID, "will not be found", None, None, AttributeError, "No such stream "),
        (COUNT, None, None, None, None, None, ValueError, "No catalog defined.  Multiple catalog "),
        (COUNT, None, None, "baseline", None, None, ValueError, "No catalog defined.  Multiple catalog "),
        (COUNT, "will not be found", None, "baseline", None, None, ValueError, "No catalog defined.  Multiple catalog "),
        (COUNT, "aps", None, None, None, None, ValueError, "No catalog defined.  Multiple catalog "),
        (COUNT, None, None, "primary", None, None, ValueError, "No catalog defined.  Multiple catalog "),
        (COUNT, "aps", None, "primary", None, None, ValueError, "No catalog defined.  Multiple catalog "),
    ],
)
def test_utils_getStreamValues_Exceptionr(
    scan_id, key, db, stream, cat, query, v1, error, first_words
):
    """Look for known errors from getStreamValues()."""
    if db == DEFAULT_CATALOG_ID:
        db = cat
    with pytest.raises(error) as exc:
        utils.getStreamValues(
            scan_id, db=db, key_fragment=key, stream=stream, query=query, use_v1=v1
        )
    assert str(exc.value).startswith(first_words)
# fmt: on



"""
Ophyd support for the EPICS scalcout record

http://htmlpreview.github.io/?https://github.com/epics-modules/calc/blob/R3-6-1/documentation/calcDocs.html

Public Structures

.. autosummary::

    ~UserScalcoutDevice
    ~UserScalcoutN
    ~ScalcoutRecord
    ~ScalcoutRecordNumberChannel
    ~ScalcoutRecordStringChannel
"""

from collections import OrderedDict
from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as DDC
from ophyd import EpicsSignal
from ophyd import FormattedComponent as FC

from ._common import EpicsRecordDeviceCommonAll
from ._common import EpicsRecordFloatFields
from ._common import EpicsSynAppsRecordEnableMixin
from .. import utils as APS_utils


CHANNEL_LETTERS_LIST = "A B C D E F G H I J K L".split()


class ScalcoutRecordNumberChannel(Device):
    """
    Number channel of an scalcout record: A-L

    .. index:: Ophyd Device; synApps ScalcoutRecordNumberChannel

    .. autosummary::

        ~reset
    """

    input_value = FC(EpicsSignal, "{prefix}.{_ch}", kind="config")
    input_pv = FC(EpicsSignal, "{prefix}.INP{_ch}", kind="config")

    def __init__(self, prefix, letter, **kwargs):
        self._ch = letter
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.input_pv.put("")
        self.input_value.put(0)


class ScalcoutRecordStringChannel(Device):
    """
    String channel of an scalcout record: AA-LL

    .. index:: Ophyd Device; synApps ScalcoutRecordStringChannel

    .. autosummary::

        ~reset
    """

    input_value = FC(EpicsSignal, "{prefix}.{_ch}", kind="config")
    input_pv = FC(EpicsSignal, "{prefix}.IN{_ch}", kind="config")

    def __init__(self, prefix, letter, **kwargs):
        self._ch = letter
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.input_pv.put("")
        self.input_value.put("")


def _channels(input_list):
    defn = OrderedDict()
    for nsym in input_list:
        defn[nsym] = (ScalcoutRecordNumberChannel, "", {"letter": nsym})
        ssym = nsym + nsym
        defn[ssym] = (ScalcoutRecordStringChannel, "", {"letter": ssym})

    return defn


class ScalcoutRecord(EpicsRecordFloatFields, EpicsRecordDeviceCommonAll):
    """
    EPICS SynApps calc scalcout record support in ophyd

    .. index:: Ophyd Device; synAppsScalcoutRecord

    .. autosummary::

        ~reset

    :see: http://htmlpreview.github.io/?https://github.com/epics-modules/calc/blob/R3-6-1/documentation/calcDocs.html
    """
    units = Cpt(EpicsSignal, ".EGU", kind="config")
    precision = Cpt(EpicsSignal, ".PREC", kind="config")

    calculation = Cpt(EpicsSignal, ".CALC", kind="config", string=True)
    calculated_value = Cpt(EpicsSignal, ".VAL", kind="hinted")
    calculated_value_string = Cpt(EpicsSignal, ".SVAL", kind="normal")

    output_calculation = Cpt(EpicsSignal, ".OCAL", kind="config", string=True)
    output_value = Cpt(EpicsSignal, ".OVAL", kind="config")
    output_value_string = Cpt(EpicsSignal, ".OSV", kind="hinted", string=True)
    output_pv = Cpt(EpicsSignal, ".OUT", kind="config")
    output_execute_option = Cpt(EpicsSignal, ".OOPT", kind="config")
    output_execution_delay = Cpt(EpicsSignal, ".ODLY", kind="config")
    output_data_option = Cpt(EpicsSignal, ".DOPT", kind="config")
    invalid_output_action = Cpt(EpicsSignal, ".IVOA", kind="config")
    invalid_output_value = Cpt(EpicsSignal, ".IVOV", kind="config")
    event_to_issue = Cpt(EpicsSignal, ".OEVT", kind="config")

    output_pv_status = Cpt(EpicsSignal, ".OUTV", kind="config")
    calculation_valid = Cpt(EpicsSignal, ".CLCV", kind="config")
    output_calculation_valid = Cpt(EpicsSignal, ".OCLV", kind="config")
    output_delay_active = Cpt(EpicsSignal, ".DLYA", kind="config")
    wait_for_completion = Cpt(EpicsSignal, ".WAIT", kind="config")

    channels = DDC(_channels(CHANNEL_LETTERS_LIST))

    read_attrs = (
        APS_utils.itemizer("channels.%s", CHANNEL_LETTERS_LIST)
        +
        APS_utils.itemizer("channels.%s", [c+c for c in CHANNEL_LETTERS_LIST])
    )
    hints = {"fields": read_attrs}

    @property
    def value(self):
        return self.calculated_value.get()

    def reset(self):
        """set all fields to default values"""
        pvname = self.description.pvname.split(".")[0]
        self.enable.put("E")
        self.scanning_rate.put("Passive")
        self.description.put(pvname)
        self.units.put("")
        self.precision.put("5")

        self.calculation.put("0")
        self.calculated_value.put(0)
        self.output_calculation.put("")
        self.output_value.put(0)

        self.forward_link.put("")
        self.output_pv.put("")
        self.invalid_output_action.put(0)
        self.invalid_output_value.put(0)

        self.output_execution_delay.put(0)
        self.output_execute_option.put(0)
        self.output_data_option.put(0)

        for letter in self.channels.read_attrs:
            channel = getattr(self.channels, letter)
            if isinstance(
                channel,
                (ScalcoutRecordNumberChannel, ScalcoutRecordStringChannel)
            ):
                channel.reset()
        self.read_attrs = (
            ["channels.%s" % c for c in CHANNEL_LETTERS_LIST]
            +
            ["channels.%s" % c+c for c in CHANNEL_LETTERS_LIST]
        )
        self.hints = {"fields": self.read_attrs}


class UserScalcoutN(EpicsSynAppsRecordEnableMixin, ScalcoutRecord):
    """Single instance of the userStringCalcN database."""


class UserScalcoutDevice(Device):
    """
    EPICS synApps XXX IOC setup of user scalcouts: ``$(P):userStringCalc$(N)``

    .. index:: Ophyd Device; synApps UserScalcoutDevice

    .. autosummary::

        ~reset

    """

    enable = Cpt(EpicsSignal, "userStringCalcEnable", kind="omitted")
    scalcout1 = Cpt(UserScalcoutN, "userStringCalc1")
    scalcout2 = Cpt(UserScalcoutN, "userStringCalc2")
    scalcout3 = Cpt(UserScalcoutN, "userStringCalc3")
    scalcout4 = Cpt(UserScalcoutN, "userStringCalc4")
    scalcout5 = Cpt(UserScalcoutN, "userStringCalc5")
    scalcout6 = Cpt(UserScalcoutN, "userStringCalc6")
    scalcout7 = Cpt(UserScalcoutN, "userStringCalc7")
    scalcout8 = Cpt(UserScalcoutN, "userStringCalc8")
    scalcout9 = Cpt(UserScalcoutN, "userStringCalc9")
    scalcout10 = Cpt(UserScalcoutN, "userStringCalc10")

    def reset(self):  # lgtm [py/similar-function]
        """set all fields to default values"""
        self.scalcout1.reset()
        self.scalcout2.reset()
        self.scalcout3.reset()
        self.scalcout4.reset()
        self.scalcout5.reset()
        self.scalcout6.reset()
        self.scalcout7.reset()
        self.scalcout8.reset()
        self.scalcout9.reset()
        self.scalcout10.reset()
        self.read_attrs = ["scalcout%d" % (c + 1) for c in range(10)]
        self.read_attrs.insert(0, "enable")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
NeXus File Writer Callbacks
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~NXWriter
   ~NXWriterAPS
"""

import datetime
import h5py
import logging
import numpy as np
import os
import yaml

from .callback_base import FileWriterCallbackBase


NEXUS_FILE_EXTENSION = "hdf"  # use this file extension for the output
NEXUS_RELEASE = "v2020.1"  # NeXus release to which this file is written
logger = logging.getLogger(__name__)


class NXWriter(FileWriterCallbackBase):
    """
    General class for writing HDF5/NeXus file (using only NeXus base classes).

    .. index:: Bluesky Callback; NXWriter

    New with apstools release 1.3.0.

    One scan is written to one HDF5/NeXus file.

    METHODS

    .. autosummary::

       ~writer
       ~h5string
       ~add_dataset_attributes
       ~assign_signal_type
       ~create_NX_group
       ~get_sample_title
       ~get_stream_link
       ~write_data
       ~write_detector
       ~write_entry
       ~write_instrument
       ~write_metadata
       ~write_monochromator
       ~write_positioner
       ~write_root
       ~write_sample
       ~write_slits
       ~write_source
       ~write_streams
       ~write_user
    """

    warn_on_missing_content = True
    file_extension = NEXUS_FILE_EXTENSION
    instrument_name = None  # name of this instrument
    # NeXus release to which this file is written
    nexus_release = NEXUS_RELEASE
    nxdata_signal = None  # name of dataset for Y axis on plot
    nxdata_signal_axes = None  # name of dataset for X axis on plot
    root = None  # instance of h5py.File

    # convention: methods written in alphabetical order

    def add_dataset_attributes(self, ds, v, long_name=None):
        """
        add attributes from v dictionary to dataset ds
        """
        ds.attrs["units"] = self.h5string(v["units"])
        ds.attrs["source"] = self.h5string(v["source"])
        if long_name is not None:
            ds.attrs["long_name"] = self.h5string(long_name)
        if v["dtype"] not in ("string",):
            ds.attrs["precision"] = v["precision"]

            def cautious_set(key):
                if v[key] is not None:
                    ds.attrs[key] = v[key]

            cautious_set("lower_ctrl_limit")
            cautious_set("upper_ctrl_limit")

    def assign_signal_type(self):
        """
        decide if a signal in the primary stream is a detector or a positioner
        """
        try:
            primary = self.root[
                "/entry/instrument/bluesky/streams/primary"
            ]
        except KeyError:
            raise KeyError(
                f"no primary data stream in "
                f"scan {self.scan_id} ({self.uid[:7]})"
            )
        for k, v in primary.items():
            # logger.debug(v.name)
            # logger.debug(v.keys())
            if k in self.detectors:
                signal_type = "detector"
            elif k in self.positioners:
                signal_type = "positioner"
            else:
                signal_type = "other"
            v.attrs["signal_type"] = signal_type  # group
            try:
                v["value"].attrs["signal_type"] = signal_type  # dataset
            except KeyError:
                if self.warn_on_missing_content:
                    logger.warning(
                        "Could not assign %s as signal type %s", k, signal_type
                    )

    def create_NX_group(self, parent, specification):
        """
        create an h5 group with named NeXus class (specification)
        """
        local_address, nx_class = specification.split(":")
        if not nx_class.startswith("NX"):
            raise ValueError(
                "NeXus base class must start with 'NX',"
                f" received {nx_class}"
            )
        group = parent.create_group(local_address)
        group.attrs["NX_class"] = nx_class
        group.attrs["target"] = group.name  # for use as NeXus link
        return group

    def getResourceFile(self, resource_id):
        """
        full path to the resource file specified by uid ``resource_id``

        override in subclass as needed
        """
        # reject unsupported specifications
        resource = self.externals[resource_id]
        if resource["spec"] not in ("AD_HDF5",):
            # HDF5-specific implementation for now
            raise ValueError(
                f'{resource_id}: spec {resource["spec"]} not handled'
            )

        # logger.debug(yaml.dump(resource))
        fname = os.path.join(resource["root"], resource["resource_path"],)
        return fname

    def get_sample_title(self):
        """
        return the title for this sample

        default title: {plan_name}-S{scan_id}-{short_uid}
        """
        return f"{self.plan_name}-S{self.scan_id:04d}-{self.uid[:7]}"

    def get_stream_link(self, signal, stream=None, ref=None):
        """
        return the h5 object for ``signal``

        DEFAULTS

        ``stream`` : ``baseline``
        ``key`` : ``value_start``
        """
        stream = stream or "baseline"
        ref = ref or "value_start"
        h5_addr = (
            f"/entry/instrument/bluesky/streams/{stream}/{signal}/{ref}"
        )
        if h5_addr not in self.root:
            raise KeyError(f"HDF5 address {h5_addr} not found.")
        # return the h5 object, to make a link
        return self.root[h5_addr]

    def h5string(self, text):
        """Format string for h5py interface."""
        if isinstance(text, (tuple, list)):
            return [self.h5string(str(t)) for t in text]
        text = text or ""
        return text.encode("utf8")

    def writer(self):
        """
        write collected data to HDF5/NeXus data file
        """
        fname = self.file_name or self.make_file_name()
        with h5py.File(fname, "w") as self.root:
            self.write_root(fname)

        self.root = None
        logger.info(f"wrote NeXus file: {fname}")  # lgtm [py/clear-text-logging-sensitive-data]
        self.output_nexus_file = fname

    def write_data(self, parent):
        """
        group: /entry/data:NXdata
        """
        nxdata = self.create_NX_group(parent, "data:NXdata")

        primary = parent["instrument/bluesky/streams/primary"]
        for k in primary.keys():
            nxdata[k] = primary[k + "/value"]

        # pick the timestamps from one of the datasets (the last one)
        nxdata["EPOCH"] = primary[k + "/time"]

        signal_attribute = self.nxdata_signal
        if signal_attribute is None:
            if len(self.detectors) > 0:
                # arbitrary but consistent choice
                signal_attribute = self.detectors[0]
        axes_attribute = self.nxdata_signal_axes
        if axes_attribute is None:
            if len(self.positioners) > 0:
                # TODO: what if wrong shape here?
                axes_attribute = self.positioners

        # TODO: rabbit-hole alert!  simplify
        # this code is convoluted (like the selection logic)
        # Is there a library to help? databroker? event_model? area_detector_handlers?
        fmt = "Cannot set %s as %s attribute, no such dataset"
        if signal_attribute is not None:
            if signal_attribute in nxdata:
                nxdata.attrs["signal"] = signal_attribute

                axes = []
                for ax in axes_attribute or []:
                    if ax in nxdata:
                        axes.append(ax)
                    else:
                        if self.warn_on_missing_content:
                            logger.warning(fmt, ax, "axes")
                if axes_attribute is not None:
                    nxdata.attrs["axes"] = axes_attribute
            else:
                if self.warn_on_missing_content:
                    logger.warning(fmt, signal_attribute, "signal")

        return nxdata

    def write_detector(self, parent):
        """
        group: /entry/instrument/detectors:NXnote/DETECTOR:NXdetector
        """
        if len(self.detectors) == 0:
            logger.info("No detectors identified.")
            return

        primary = parent["/entry/instrument/bluesky/streams/primary"]

        group = self.create_NX_group(parent, "detectors:NXnote")
        for k, v in primary.items():
            if v.attrs.get("signal_type") != "detector":
                continue
            nxdetector = self.create_NX_group(group, f"{k}:NXdetector")
            nxdetector["data"] = v

        return group

    def write_entry(self):
        """
        group: /entry/data:NXentry
        """
        nxentry = self.create_NX_group(
            self.root, self.root.attrs["default"] + ":NXentry"
        )

        nxentry.create_dataset(
            "start_time",
            data=datetime.datetime.fromtimestamp(
                self.start_time
            ).isoformat(),
        )
        nxentry.create_dataset(
            "end_time",
            data=datetime.datetime.fromtimestamp(
                self.stop_time
            ).isoformat(),
        )
        ds = nxentry.create_dataset(
            "duration", data=self.stop_time - self.start_time
        )
        ds.attrs["units"] = "s"

        nxentry.create_dataset("program_name", data="bluesky")

        self.write_instrument(nxentry)  # also writes streams and metadata
        try:
            nxdata = self.write_data(nxentry)
            nxentry.attrs["default"] = nxdata.name.split("/")[-1]
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning(exc)

        self.write_sample(nxentry)
        self.write_user(nxentry)

        # apply links
        h5_addr = "/entry/instrument/source/cycle"
        if h5_addr in self.root:
            nxentry["run_cycle"] = self.root[h5_addr]
        else:
            if self.warn_on_missing_content:
                logger.warning("No data for /entry/run_cycle")

        nxentry["title"] = self.get_sample_title()
        nxentry["plan_name"] = self.root[
            "/entry/instrument/bluesky/metadata/plan_name"
        ]
        nxentry["entry_identifier"] = self.root[
            "/entry/instrument/bluesky/uid"
        ]

        return nxentry

    def write_instrument(self, parent):
        """
        group: /entry/instrument:NXinstrument
        """
        nxinstrument = self.create_NX_group(
            parent, "instrument:NXinstrument"
        )
        bluesky_group = self.create_NX_group(
            nxinstrument, "bluesky:NXnote"
        )

        md_group = self.write_metadata(bluesky_group)
        self.write_streams(bluesky_group)

        bluesky_group["uid"] = md_group["run_start_uid"]
        bluesky_group["plan_name"] = md_group["plan_name"]

        try:
            self.assign_signal_type()
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning(exc)

        self.write_slits(nxinstrument)
        try:
            self.write_detector(nxinstrument)
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning(exc)

        self.write_monochromator(nxinstrument)
        try:
            self.write_positioner(nxinstrument)
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning(exc)
        self.write_source(nxinstrument)
        return nxinstrument

    def write_metadata(self, parent):
        """
        group: /entry/instrument/bluesky/metadata:NXnote

        metadata from the bluesky start document
        """
        group = self.create_NX_group(parent, "metadata:NXnote")

        ds = group.create_dataset("run_start_uid", data=self.uid)
        ds.attrs["long_name"] = "bluesky run uid"
        ds.attrs["target"] = ds.name

        for k, v in self.metadata.items():
            is_yaml = False
            if isinstance(v, (dict, tuple, list)):
                # fallback technique: save complicated structures as YAML text
                v = yaml.dump(v)
                is_yaml = True
            if isinstance(v, str):
                v = self.h5string(v)
            ds = group.create_dataset(k, data=v)
            ds.attrs["target"] = ds.name
            if is_yaml:
                ds.attrs["text_format"] = "yaml"

        return group

    def write_monochromator(self, parent):
        """
        group: /entry/instrument/monochromator:NXmonochromator
        """
        pre = "monochromator_dcm"
        keys = "wavelength energy theta y_offset mode".split()

        try:
            links = {
                key: self.get_stream_link(f"{pre}_{key}") for key in keys
            }
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning(
                    "%s -- not creating monochromator group", str(exc)
                )
            return

        pre = "monochromator"
        key = "feedback_on"
        try:
            links[key] = self.get_stream_link(f"{pre}_{key}")
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning("%s -- feedback signal not found", str(exc))

        nxmonochromator = self.create_NX_group(
            parent, "monochromator:NXmonochromator"
        )
        for k, v in links.items():
            nxmonochromator[k] = v
        return nxmonochromator

    def write_positioner(self, parent):
        """
        group: /entry/instrument/positioners:NXnote/POSITIONER:NXpositioner
        """
        if len(self.positioners) == 0:
            logger.info("No positioners identified.")
            return

        primary = parent["/entry/instrument/bluesky/streams/primary"]

        group = self.create_NX_group(parent, "positioners:NXnote")
        for k, v in primary.items():
            if v.attrs.get("signal_type") != "positioner":
                continue
            nxpositioner = self.create_NX_group(group, f"{k}:NXpositioner")
            nxpositioner["value"] = v

        return group

    def write_root(self, filename):
        """
        root of the HDF5 file
        """
        self.root.attrs["file_name"] = filename
        self.root.attrs["file_time"] = datetime.datetime.now().isoformat()
        if self.instrument_name is not None:
            self.root.attrs["instrument"] = self.instrument_name
        self.root.attrs["creator"] = self.__class__.__name__
        self.root.attrs["NeXus_version"] = self.nexus_release
        self.root.attrs["HDF5_Version"] = h5py.version.hdf5_version
        self.root.attrs["h5py_version"] = h5py.version.version
        self.root.attrs["default"] = "entry"

        self.write_entry()

    def write_sample(self, parent):
        """
        group: /entry/sample:NXsample
        """
        pre = "sample_data"
        keys = """
            chemical_formula
            concentration
            description
            electric_field
            magnetic_field
            rotation_angle
            scattering_length_density
            stress_field
            temperature
            volume_fraction
            x_translation
        """.split()

        links = {}
        for key in keys:
            try:
                links[key] = self.get_stream_link(f"{pre}_{key}")
            except KeyError as exc:
                if self.warn_on_missing_content:
                    logger.warning("%s", str(exc))
        if len(links) == 0:
            if self.warn_on_missing_content:
                logger.warning(
                    "no sample data found, not creating sample group"
                )
            return

        nxsample = self.create_NX_group(parent, "sample:NXsample")
        for k, v in links.items():
            nxsample[k] = v

        for key in "electric_field magnetic_field stress_field".split():
            ds = nxsample[key]
            ds.attrs["direction"] = self.get_stream_link(
                f"{pre}_{key}_dir"
            )[()].lower()

        return nxsample

    def write_slits(self, parent):
        """
        group: /entry/instrument/slits:NXnote/SLIT:NXslit

        override in subclass to store content, name patterns
        vary with each instrument
        """
        # group = self.create_NX_group(parent, f"slits:NXnote")

    def write_source(self, parent):
        """
        group: /entry/instrument/source:NXsource

        Note: this is (somewhat) generic, override for a different source
        """
        nxsource = self.create_NX_group(parent, "source:NXsource")

        ds = nxsource.create_dataset("name", data="Bluesky framework")
        ds.attrs["short_name"] = "bluesky"
        nxsource.create_dataset("type", data="Synchrotron X-ray Source")
        nxsource.create_dataset("probe", data="x-ray")

        return nxsource

    def write_stream_external(
        self, parent, d, subgroup, stream_name, k, v
    ):
        # TODO: rabbit-hole alert! simplify
        # lots of variations possible

        # count number of unique resources (expect only 1)
        resource_id_list = []
        for datum_id in d:
            resource_id = self.externals[datum_id]["resource"]
            if resource_id not in resource_id_list:
                resource_id_list.append(resource_id)
        if len(resource_id_list) != 1:
            raise ValueError(
                f"{len(resource_id_list)}"
                f" unique resource UIDs: {resource_id_list}"
            )

        fname = self.getResourceFile(resource_id)
        logger.info("reading %s from EPICS AD data file: %s", k, fname)
        with h5py.File(fname, "r") as hdf_image_file_root:
            h5_obj = hdf_image_file_root["/entry/data/data"]
            ds = subgroup.create_dataset(
                "value",
                data=h5_obj[()],
                compression="lzf",
                # compression="gzip",
                # compression_opts=9,
                shuffle=True,
                fletcher32=True,
            )
            ds.attrs["target"] = ds.name
            ds.attrs["source_file"] = fname
            ds.attrs["source_address"] = h5_obj.name
            ds.attrs["resource_id"] = resource_id
            ds.attrs["units"] = ""

        subgroup.attrs["signal"] = "value"

    def write_stream_internal(
        self, parent, d, subgroup, stream_name, k, v
    ):
        # fmt: off
        subgroup.attrs["signal"] = "value"
        subgroup.attrs["axes"] = ["time", ]
        # fmt: on
        if isinstance(d, list) and len(d) > 0:
            if v["dtype"] in ("string",):
                d = self.h5string(d)
            elif v["dtype"] in ("integer", "number"):
                d = np.array(d)
        try:
            ds = subgroup.create_dataset("value", data=d)
            ds.attrs["target"] = ds.name
            try:
                self.add_dataset_attributes(ds, v, k)
            except Exception as exc:
                logger.error("%s %s %s %s", v["dtype"], type(d), k, exc)
        except TypeError as exc:
            logger.error(
                "%s %s %s %s",
                v["dtype"],
                k,
                f"TypeError({exc})",
                v["data"],
            )
        if stream_name == "baseline":
            # make it easier to pick single values
            # identify start/end of acquisition
            ds = subgroup.create_dataset("value_start", data=d[0])
            self.add_dataset_attributes(ds, v, k)
            ds.attrs["target"] = ds.name
            ds = subgroup.create_dataset("value_end", data=d[-1])
            self.add_dataset_attributes(ds, v, k)
            ds.attrs["target"] = ds.name

    def write_streams(self, parent):
        """
        group: /entry/instrument/bluesky/streams:NXnote

        data from all the bluesky streams
        """
        bluesky = self.create_NX_group(parent, "streams:NXnote")
        for stream_name, uids in self.streams.items():
            if len(uids) != 1:
                raise ValueError(
                    f"stream {len(uids)} has descriptors, expecting only 1"
                )
            group = self.create_NX_group(bluesky, stream_name + ":NXnote")
            uid0 = uids[0]  # just get the one descriptor uid
            group.attrs["uid"] = uid0
            # just get the one descriptor
            acquisition = self.acquisitions[uid0]
            for k, v in acquisition["data"].items():
                d = v["data"]
                # NXlog is for time series data but NXdata makes an automatic plot
                subgroup = self.create_NX_group(group, k + ":NXdata")

                if v["external"]:
                    self.write_stream_external(
                        parent, d, subgroup, stream_name, k, v
                    )
                else:
                    self.write_stream_internal(
                        parent, d, subgroup, stream_name, k, v
                    )

                t = np.array(v["time"])
                ds = subgroup.create_dataset("EPOCH", data=t)
                ds.attrs["units"] = "s"
                ds.attrs["long_name"] = "epoch time (s)"
                ds.attrs["target"] = ds.name

                t_start = t[0]
                ds = subgroup.create_dataset("time", data=t - t_start)
                ds.attrs["units"] = "s"
                ds.attrs["long_name"] = "time since first data (s)"
                ds.attrs["target"] = ds.name
                ds.attrs["start_time"] = t_start
                ds.attrs[
                    "start_time_iso"
                ] = datetime.datetime.fromtimestamp(t_start).isoformat()

            # link images to parent names
            for k in group:
                if k.endswith("_image") and k[:-6] not in group:
                    group[k[:-6]] = group[k]

        return bluesky

    def write_user(self, parent):
        """
        group: /entry/contact:NXuser
        """
        keymap = dict(
            name="bss_user_info_contact",
            affiliation="bss_user_info_institution",
            email="bss_user_info_email",
            facility_user_id="bss_user_info_badge",
        )

        try:
            links = {k: self.get_stream_link(v) for k, v in keymap.items()}
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning("%s -- not creating source group", str(exc))
            return

        nxuser = self.create_NX_group(parent, "contact:NXuser")
        nxuser.create_dataset("role", data="contact")
        for k, v in links.items():
            nxuser[k] = v
        return nxuser


class NXWriterAPS(NXWriter):
    """
    Customize :class:`~NXWriter` with APS-specific content.

    .. index:: Bluesky Callback; NXWriterAPS

    New with apstools release 1.3.0.

    * Adds `/entry/instrument/undulator` group if metadata exists.
    * Adds APS information to `/entry/instrument/source` group.

    .. autosummary::

       ~write_instrument
       ~write_source
       ~write_undulator
    """

    # convention: methods written in alphabetical order

    def write_instrument(self, parent):
        """
        group: /entry/instrument:NXinstrument
        """
        nxinstrument = super().write_instrument(parent)
        self.write_undulator(nxinstrument)

    def write_source(self, parent):
        """
        group: /entry/instrument/source:NXsource

        Note: this is specific to the APS, override for a different source
        """
        pre = "aps"
        keys = "current fill_number".split()

        try:
            # fmt: off
            links = {
                key: self.get_stream_link(f"{pre}_{key}")
                for key in keys
            }
            # fmt: on
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning("%s -- not creating source group", str(exc))
            return

        nxsource = self.create_NX_group(parent, "source:NXsource")
        for k, v in links.items():
            nxsource[k] = v

        ds = nxsource.create_dataset("name", data="Advanced Photon Source")
        ds.attrs["short_name"] = "APS"
        nxsource.create_dataset("type", data="Synchrotron X-ray Source")
        nxsource.create_dataset("probe", data="x-ray")
        ds = nxsource.create_dataset("energy", data=6)
        ds.attrs["units"] = "GeV"

        try:
            nxsource["cycle"] = self.get_stream_link("aps_aps_cycle")
        except KeyError:
            pass  # Should we compute the cycle?

        return nxsource

    def write_undulator(self, parent):
        """
        group: /entry/instrument/undulator:NXinsertion_device
        """
        pre = "undulator_downstream"
        keys = """
            device location
            energy
            energy_taper
            gap
            gap_taper
            harmonic_value
            total_power
            version
        """.split()

        try:
            links = {
                key: self.get_stream_link(f"{pre}_{key}") for key in keys
            }
        except KeyError as exc:
            if self.warn_on_missing_content:
                logger.warning("%s -- not creating undulator group", str(exc))
            return

        undulator = self.create_NX_group(
            parent, "undulator:NXinsertion_device"
        )
        undulator.create_dataset("type", data="undulator")
        for k, v in links.items():
            undulator[k] = v
        return undulator

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Support for logging
+++++++++++++++++++++++++++++++++++++++

There is a guide describing :ref:`howto_session_logs`.

.. autosummary::

   ~file_log_handler
   ~get_log_path
   ~setup_IPython_console_logging
   ~stream_log_handler
"""

import logging
import pathlib


def file_log_handler(
    file_name_base,
    maxBytes=0,
    backupCount=0,
    log_path=None,
    level=None,
):
    """
    Record logging output to a file.

    PARAMETERS

    file_name_base : *str*
        Part of the name to store the log file. Full name is
        ``f"<log_path>/{file_name_base}.log"`` in present working directory.
    log_path : *str*
        Part of the name to store the log file. Full name is
        ``f"<log_path>/{file_name_base}.log"`` in present working directory.
        default: (the present working directory)/LOG_DIR_BASE
    level : *int*
        Threshold for reporting messages with this logger. Logging messages
        which are less severe than ``level`` will be ignored.
        default: 10 (``logging.DEBUG`` or ``DEBUG``)
        see: https://docs.python.org/3/library/logging.html#levels
    maxBytes : (optional) *int*
        Log file *rollover* begins whenever the current log file is nearly
        *maxBytes* in length.  A new file is written when the current line will
        push the current file beyond this limit.
        default: 0
    backupCount : (optional) *int*
        When *backupCount* is non-zero, the system will keep up to *backupCount*
        numbered log files (with added extensions `.1`, '.2`, ...).  The current
        log file always has no numbered extension.  The previous log file is the
        one with the lowest extension number.
        default: 0

    .. note::  When either ``maxBytes`` or ``backupCount`` are zero,
        log file rollover never occurs, so you generally want to set
        ``backupCount`` to at least 1, and have a non-zero ``maxBytes``.
    """
    log_path = log_path or get_log_path()
    log_file = log_path / f"{file_name_base}.log"
    level = level or logging.DEBUG

    if maxBytes > 0 or backupCount > 0:
        handler = logging.handlers.RotatingFileHandler(log_file, maxBytes=maxBytes, backupCount=backupCount)
    else:
        handler = logging.FileHandler(log_file)

    handler.setLevel(level)

    formatter = logging.Formatter(
        (
            "|%(asctime)s"
            "|%(levelname)s"
            "|%(process)d"
            "|%(name)s"
            "|%(module)s"
            "|%(lineno)d"
            "|%(threadName)s"
            "| - "
            "%(message)s"
        )
    )
    formatter.default_msec_format = "%s.%03d"
    handler.setFormatter(formatter)

    return handler


def get_log_path():
    """
    Return a path to `./.logs`. Create directory if it does not exist.
    """
    path = pathlib.Path().cwd() / ".logs"
    if not path.exists():
        path.mkdir()
    return path


def setup_IPython_console_logging(logger=None, filename="ipython_console.log"):
    """
    Record all input (``In``) and output (``Out``) from IPython console.

    PARAMETERS

    logger
        *object*:
        Instance of ``logging.Logger``.
    filename
        *str*:
        Name of the log file.
        (default: ``ipython_console.log``)
    """
    try:
        from IPython import get_ipython

        # start logging console to file
        # https://ipython.org/ipython-doc/3/interactive/magics.html#magic-logstart
        _ipython = get_ipython()
        io_file = get_log_path() / filename
        if _ipython is not None:
            _ipython.magic(f"logstart -o -t {io_file} rotate")
        if logger is not None:
            logger.info("Console logging: %s", io_file)
    except Exception as exc:
        if logger is not None:
            logger.exception("Could not setup console logging.")
        else:
            print(f"Could not setup console logging: {exc}")


def stream_log_handler(formatter=None, level="INFO"):
    """
    Record logging output to a stream (such as the console).

    PARAMETERS

    formatter
        *object*:
        Instance of ``logging.Formatter``.
    level
        *str*:
        Name of the logging level to report.
        (default: ``INFO``)
    """
    handler = logging.StreamHandler()

    # nice output format
    # https://docs.python.org/3/library/logging.html#logrecord-attributes
    if formatter is None:
        # fmt: off
        formatter = logging.Formatter(
            (
                "%(levelname)-.1s"		# only first letter
                " %(asctime)s"
                " - "
                "%(message)s"
            ),
            datefmt="%a-%H:%M:%S"
        )
        # fmt: on
        formatter.default_msec_format = "%s.%03d"
    handler.setFormatter(formatter)
    handler.setLevel(level)

    return handler



from ..utils import run_in_thread
import bluesky.plan_stubs as bps
import logging

logger = logging.getLogger(__name__)
POLL_DELAY = 0.000_05


def run_blocking_function(function, *args, **kwargs):
    """
    Run a blocking function as a bluesky plan, in a thread.

    The ``run_blocking_function()`` is a bluesky plan which
    runs ``function(*args, **kwargs)`` in a thread so it does
    not block the RunEngine's background operations.

    .. note: Any result(s) returned from ``function``
       will be ignored.

    It is intended to call blocking code that should not be
    called directly from a bluesky plan.

    USAGE::

        yield from run_blocking_function(function, *args, **kwargs)

    EXAMPLE:

    This example creates a bluesky plan named ``start_incrementer()``
    which configures a synApps "userCalc" as an automated
    10Hz incrementer.
    Configuration is made by calling the *blocking* function
    :func:`~apstools.synApps.swait.setup_incrementer_swait`.
    The incrementer resets to zero at 100,000:

    .. code-block:: python

        from apstools.plans import run_blocking_function
        from apstools.synApps import SwaitRecord
        from apstools.synApps import setup_incrementer_swait
        import bluesky.plan_stubs as bps

        swait = SwaitRecord("ioc:userCalc1", name="swait")

        def start_incrementer():
            yield from run_blocking_function(
                setup_incrementer_swait,
                swait,
                scan="0.1 second",
                limit=100_000
            )

        # now, run my_plan
        RE(my_plan())

    .. note: This function is a convenience adapter to execute
       functions that are not bluesky generator functions, plans.
       It's best not to use this as a decorator.

    (new in release 1.6.3)
    """

    @run_in_thread
    def internal():
        result = function(*args, **kwargs)
        logger.debug("%s() result=%s", result)

    thread = internal()
    while thread.is_alive():
        yield from bps.sleep(POLL_DELAY)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Synthetic pseudo-Voigt function
+++++++++++++++++++++++++++++++++++++++

EXAMPLES:

.. code-block:: python
    :caption: Simple example of SynPseudoVoigt().
    :linenos:

    from apstools.devices import SynPseudoVoigt
    from ophyd.sim import motor
    det = SynPseudoVoigt('det', motor, 'motor',
        center=0, eta=0.5, scale=1, sigma=1, bkg=0)

    # scan the "det" peak with the "motor" positioner
    # RE(bp.scan([det], motor, -2, 2, 41))


.. code-block:: python
    :caption: Example of SynPseudoVoigt() with randomized values.
    :linenos:

    import numpy as np
    from apstools.devices import SynPseudoVoigt
    synthetic_pseudovoigt = SynPseudoVoigt(
        'synthetic_pseudovoigt', m1, 'm1',
        center=-1.5 + 0.5*np.random.uniform(),
        eta=0.2 + 0.5*np.random.uniform(),
        sigma=0.001 + 0.05*np.random.uniform(),
        scale=1e5,
        bkg=0.01*np.random.uniform())

    # scan the "synthetic_pseudovoigt" peak with the "m1" positioner
    # RE(bp.scan([synthetic_pseudovoigt], m1, -2, 0, 219))

.. autosummary::

   ~SynPseudoVoigt
"""

import ophyd.sim
import numpy as np


class SynPseudoVoigt(ophyd.sim.SynSignal):  # lgtm [py/missing-call-to-init]
    """
    Evaluate a point on a pseudo-Voigt based on the value of a motor.

    .. index:: Ophyd Signal; SynPseudoVoigt

    Provides a signal to be measured.
    Acts like a detector.

    :see: https://en.wikipedia.org/wiki/Voigt_profile

    PARAMETERS

    name *str* :
        name of detector signal
    motor positioner :
        The independent coordinate
    motor_field *str* :
        name of `motor`
    center *float* :
        (optional)
        location of maximum value, default=0
    eta *float* :
        (optional)
        0 <= eta < 1.0: Lorentzian fraction, default=0.5
    scale *float* :
        (optional)
        scale >= 1 : scale factor, default=1
    sigma *float* :
        (optional)
        sigma > 0 : width, default=1
    bkg *float* :
        (optional)
        bkg >= 0 : constant background, default=0
    noise ``"poisson"`` or ``"uniform"`` or ``None`` :
        Add noise to the result.
    noise_multiplier *float* :
        Only relevant for 'uniform' noise. Multiply the random amount of
        noise by 'noise_multiplier'
    """

    def __init__(
        # fmt: off
        self,
        name, motor, motor_field,
        center=0, eta=0.5, scale=1, sigma=1, bkg=0,
        noise=None, noise_multiplier=1,
        **kwargs
        # fmt: on
    ):
        if eta < 0.0 or eta > 1.0:
            raise ValueError("eta={} must be between 0 and 1".format(eta))
        if scale < 1.0:
            raise ValueError("scale must be >= 1")
        if sigma <= 0.0:
            raise ValueError("sigma must be > 0")
        if bkg < 0.0:
            raise ValueError("bkg must be >= 0")

        # remember these terms for later access by user
        self.name = name
        self.motor = motor
        self.motor_field = motor_field
        self.center = center
        self.eta = eta
        self.scale = scale
        self.sigma = sigma
        self.bkg = bkg
        self.noise = noise
        self.noise_multiplier = noise_multiplier

        def f_lorentzian(x, gamma):
            # return gamma / np.pi / (x**2 + gamma**2)
            return 1 / np.pi / gamma / (1 + (x / gamma) ** 2)

        def f_gaussian(x, sigma):
            numerator = np.exp(-0.5 * (x / sigma) ** 2)
            denominator = sigma * np.sqrt(2 * np.pi)
            return numerator / denominator

        def pvoigt():
            m = motor.read()[self.motor_field]["value"]
            g_max = f_gaussian(0, self.sigma)  # peak normalization
            l_max = f_lorentzian(0, self.sigma)
            v = self.bkg
            if eta > 0:
                v += self.eta * f_lorentzian(m - self.center, self.sigma) / l_max
            if eta < 1:
                v += (1 - self.eta) * f_gaussian(m - self.center, self.sigma) / g_max
            v *= self.scale
            if self.noise == "poisson":
                v = int(np.random.poisson(np.round(v), 1))
            elif self.noise == "uniform":
                v += np.random.uniform(-1, 1) * self.noise_multiplier
            return v

        ophyd.sim.SynSignal.__init__(
            self, name=name, func=pvoigt, **kwargs
        )

    def randomize_parameters(self, scale=100_000, bkg=0.01):
        """
        Set random parameters. -1 <= center < 1, 0.001 <= sigma < 0.051, 95k <= scale <= 105k
        """
        self.center = -1.0 + 2 * np.random.uniform()
        self.eta = 0.2 + 0.5*np.random.uniform()
        self.sigma = 0.001 + 0.05*np.random.uniform()
        self.scale = scale * (0.95 + 0.1 * np.random.uniform())
        self.bkg = bkg*np.random.uniform()
        self.noise = "poisson"

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Documentation of batch runs
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~addDeviceDataAsStream
   ~documentation_run
"""

from bluesky import plan_stubs as bps
from ophyd import Signal

from ..utils import ipython_shell_namespace


def addDeviceDataAsStream(devices, label):
    """
    add an ophyd Device as an additional document stream

    .. index:: Bluesky Plan; addDeviceDataAsStream

    Use this within a custom plan, such as this example::

        from apstools.plans import addDeviceStream
        ...
        yield from bps.open_run()
        # ...
        yield from addDeviceDataAsStream(prescanDeviceList, "metadata_prescan")
        # ...
        yield from custom_scan_procedure()
        # ...
        yield from addDeviceDataAsStream(postscanDeviceList, "metadata_postscan")
        # ...
        yield from bps.close_run()

    """
    yield from bps.create(name=label)
    if not isinstance(devices, list):  # just in case...
        devices = [devices]
    for d in devices:
        yield from bps.read(d)
    yield from bps.save()


def documentation_run(text, stream=None, bec=None, md=None):
    """
    Save text as a bluesky run.

    PARAMETERS

    text
        *str* :
        Text to be written.
    stream
        *str* :
        document stream, default: "primary"
    bec
        *object* :
        Instance of `bluesky.BestEffortCallback`,
        default: get from IPython shell
    md
        *dict*
        (optional):
        metadata dictionary
    """
    bec = bec or ipython_shell_namespace().get("bec")
    stream = stream or "primary"

    text_signal = Signal(value=text, name="text")

    _md = dict(purpose="save text as bluesky run", plan_name="documentation_run",)
    _md.update(md or {})

    if bec is not None:
        bec.disable_plots()
        bec.disable_table()

    uid = yield from bps.open_run(md=_md)
    yield from addDeviceDataAsStream(text_signal, stream)
    yield from bps.close_run()

    if bec is not None:
        bec.enable_table()
        bec.enable_plots()

    return uid

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
